"""Utility functions to build the TCU command strings.

We discern between the following types of commands:
    - General commands;
    - M2MD commands (on a per-axis basis);
    - Thermal Monitoring System (TSM) commands;
    - HK commands.

Reference documents:
    - RD01: TCU User Manual (ARIEL-IEEC-PL-TN-002), v1.2
    - RD02: ARIEL TCU Data Handling (ARIEL-IEEC-PL-TN-007), v1.0
    - RD03: TCU code provided by Vladimiro Noce (priv. comm.)
    - RD04: ARIEL Telescope Control Unit Design Description Document (ARIEL-IEEC-PL-DD-001), v1.10
    - RD05: ARIEL TCU FW Architecture Design (ARIEL-IEEC-PL-DD-002), v1.5
"""

from enum import StrEnum
import crcmod

from egse.ariel.tcu import TcuMode
from egse.decorators import static_vars

TCU_LOGICAL_ADDRESS = "03"  # RD02 -> Fig. 10
DATA_LENGTH = "0004"  # Vladimiro's code


class PacketType(StrEnum):
    """Packet types (read/write)."""

    W = WRITE = "20"  # Write command (RD02 -> Sect. 4.1.1)
    R = READ = "40"  # Read command (RD02 -> Sect. 4.1.2)


class CommandAddress(StrEnum):
    """Identifiers of the components of the TCU the commands have to be sent to."""

    # Adopted from Vladimiro's code

    GENERAL = "0000"  # General TCU commands -> see `show_messageGEN`, where the command string is built
    M2MD_1 = "0001"  # M2MD axis-1 commands -> see `show_messageM2M`, where the command string is built
    M2MD_2 = "0002"  # M2MD axis-2 commands -> see `show_messageM2M`, where the command string is built
    M2MD_3 = "0003"  # M2MD axis-3 commands -> see `show_messageM2M`, where the command string is built
    TSM = "0004"  # TSM commands -> see `show_messageTSM`, where the command string is built
    HK = "0005"  # HK commands -> see `show_messageHK`, where the command string is built


#####################
# Command identifiers
#####################


class GeneralCommandIdentifier(StrEnum):
    """Identifiers for the general TCU commands."""

    # Adopted from Vladimiro's code

    TCU_FIRMWARE_ID = "0000"  # Read
    TCU_MODE = "0001"  # Read/Write
    TCU_STATUS = "0002"  # Read
    TCU_SIMULATED = "0003"  # Write
    RESTART_LINKS_PERIOD_LATCH = "0004"  # Write
    RESTART_LINKS_PERIOD = "0005"  # Read/Write


class M2MDCommandIdentifier(StrEnum):
    """Identifiers for the M2MD commands."""

    # Adopted from Vladimiro's code

    OPE_MNG_COMMAND = "0000"  # Write
    OPE_MNG_EVENT_CLEAR_PROTECT_FLAG = "0001"  # Write
    OPE_MNG_EVENT_CLEAR = "0002"  # Write
    OPE_MNG_STATUS = "0003"  # Read
    OPE_MNG_EVENT_REG = "0004"  # Read
    ACQ_CURR_OFF_CORR = "1000"  # Read/Write
    ACQ_CURR_GAIN_CORR = "1001"  # Read/Write
    ACQ_AXIS_A_CURR_READ = "1002"  # Read
    ACQ_AXIS_B_CURR_READ = "1003"  # Read
    ACQ_AVE_LPF_EN = "1004"  # Write
    ACQ_OVC_CFG_FILTER = "1005"  # Write
    ACQ_AVC_FILT_TIME = "1006"  # Write
    ACQ_AVERAGE_TYPE = "1007"  # Write
    ACQ_SPK_FILT_COUNTER_LIM = "1008"  # Write
    ACQ_SPK_FILT_INCR_THR = "1009"  # Write
    PROF_GEN_AXIS_STEP = "2000"  # Read/Write
    PROF_GEN_AXIS_SPEED = "2001"  # Read/Write
    PROF_GEN_AXIS_STATE_START = "2002"  # Read/Write
    SW_RS_XX_SW_RISE = "3001"  # Read
    # TODO Check whether this is correct:
    # This is how it is hard-coded in Vladimiro's code, but in his `show_messageM2M` function, he states that there's
    # an offset of 20 between SW_RS_XX_SW_RISE and SW_RS_XX_SW_FALL (where position must in 1,...,20).  That would mean
    # that for position 15, SW_RS_XX_SW_RISE is 3015, which clashes with the hard-coded value for
    # SW_RS_XX_SW_FALL.
    SW_RS_XX_SW_FALL = "3015"  # Read


class TSMCommandIdentifier(StrEnum):
    """Identifiers for the TSM commands."""

    TSM_LATCH = "0000"  # Write
    TSM_CURRENT_VALUE = "0001"  # Read/Write
    TSM_CURRENT_OFFSET = "0002"  # Read/Write
    TSM_ADC_REGISTER_LATCH = "1000"  # Write
    TSM_ADC_ID_REGISTER = "1001"  # Read
    TSM_ADC_CONFIGURATION_REGISTER = "1002"  # Read
    TSM_ADC_HPF_REGISTER = "1003"  # Read/Write
    TSM_ADC_OFC_REGISTER = "1004"  # Read/Write
    TSM_ADC_FSC_REGISTER = "1006"  # Read/Write
    TSM_ADC_COMMAND_LATCH = "1008"  # Write
    TSM_ADC_COMMAND = "1009"  # Write
    TSM_ADC_CALIBRATION = "100A"  # Write
    TSM_ADC_VALUE_XX_CURRENTN = "2000"  # Read
    TSM_ADC_VALUE_XX_BIASN = "2001"  # Read
    TSM_ADC_VALUE_XX_CURRENTP = "2002"  # Read
    TSM_ADC_VALUE_XX_BIASP = "2003"  # Read
    TSM_ACQ_COUNTER = "20C0"  # Read


class HKCommandIdentifier(StrEnum):
    """Identifiers for the HK commands."""

    # Adopted from Vladimiro's code

    VHK_PSU_VMOTOR = "0000"  # Read
    VHK_PSU_VHI = "0001"  # Read
    VHK_PSU_VLOW = "0002"  # Read
    VHK_PSU_VMEDP = "0003"  # Read
    VHK_PSU_VMEDN = "0004"  # Read
    IHK_PSU_VMEDN = "0005"  # Read
    IHK_PSU_VMEDP = "0006"  # Read
    IHK_PSU_VLOW = "0007"  # Read
    IHK_PSU_VHI = "0008"  # Read
    IHK_PSU_VMOTOR = "0009"  # Read
    THK_PSU_FIRST = "000A"  # Read
    THK_M2MD_FIRST = "000B"  # Read
    THK_PSU_SECOND = "000C"  # Read
    THK_M2MD_SECOND = "000D"  # Read
    THK_CTS_Q1 = "000E"  # Read
    THK_CTS_Q2 = "000F"  # Read
    THK_CTS_Q3 = "0010"  # Read
    THK_CTS_Q4 = "0011"  # Read
    THK_CTS_FPGA = "0012"  # Read
    THK_CTS_ADS1282 = "0013"  # Read
    VHK_THS_RET = "0014"  # Read
    HK_ACQ_COUNTER = "0015"  # Read


def get_tsm_adc_value_xx_currentn_cmd_id(probe: int) -> str:
    """Determines the command identifier for the TSM_ADC_VALUE_XX_CURRENTN for the given probe.

    Args:
        probe (int): Thermistor identifier.

    Returns:
        Command identifier for the TSM_ADC_VALUE_XX_CURRENTN for the given probe.
    """

    return get_tsm_adc_value_xx_cmd_id(TSMCommandIdentifier.TSM_ADC_VALUE_XX_CURRENTN, probe)


def get_tsm_adc_value_xx_biasn_cmd_id(probe: int) -> str:
    """Determines the command identifier for the TSM_ADC_VALUE_XX_BIASN for the given probe.

    Args:
        probe (int): Thermistor identifier.

    Returns:
        Command identifier for the TSM_ADC_VALUE_XX_BIASN for the given probe.
    """

    return get_tsm_adc_value_xx_cmd_id(TSMCommandIdentifier.TSM_ADC_VALUE_XX_BIASN, probe)


def get_tsm_adc_value_xx_currentp_cmd_id(probe: int) -> str:
    """Determines the command identifier for the TSM_ADC_VALUE_XX_CURRENTP for the given probe.

    Args:
        probe (int): Thermistor identifier.

    Returns:
        Command identifier for the TSM_ADC_VALUE_XX_CURRENTP for the given probe.
    """

    return get_tsm_adc_value_xx_cmd_id(TSMCommandIdentifier.TSM_ADC_VALUE_XX_CURRENTP, probe)


def get_tsm_adc_value_xx_biasp_cmd_id(probe: int) -> str:
    """Determines the command identifier for the TSM_ADC_VALUE_XX_BIASP for the given probe.

    Args:
        probe (int): Thermistor identifier.

    Returns:
        Command identifier for the TSM_ADC_VALUE_XX_BIASP for the given probe.
    """

    return get_tsm_adc_value_xx_cmd_id(TSMCommandIdentifier.TSM_ADC_VALUE_XX_BIASP, probe)


def get_tsm_adc_value_xx_cmd_id(tsm_cmd_id: TSMCommandIdentifier | str, probe: int):
    """Determines the command identifier for the given TSM_ADC_VALUE_XX for the given probe.

    Args:
        tsm_cmd_id (TSMCommandIdentifier | str): TSM command identifier.
        probe (int): Thermistor identifier.

    Returns:
        Command identifier for the
        TSM_ADC_VALUE_XX_CURRENTN/TSM_ADC_VALUE_XX_BIASN/TSM_ADC_VALUE_XX_CURRENTP/TSM_ADC_VALUE_XX_BIASP for the given
        probe.
    """

    # Adopted from Vladimiro's code
    return tsm_cmd_id.value[:2] + hex(probe * 4 + int(tsm_cmd_id.value[-1]))[2:].zfill(2)


def get_sw_rs_xx_sw_cmd_id(cmd_identifier: M2MDCommandIdentifier, position: int) -> str:
    """Determines the command identifier for the given SW_RS_XX_SW for the given probe.

    Args:
        cmd_identifier (M2MDCommandIdentifier): TSM command identifier.
        position (int):

    Returns:
        Command identifier for the SW_RS_XX_SW_RISE/SW_RS_XX_SW_FALL for the given probe.
    """

    # Adopted from Vladimiro's code
    return f"{cmd_identifier.value[:2]}{hex(position)[2:].zfill(2)}"


##############################
# Building the command strings
##############################


def format_value(value: int | str | StrEnum) -> str:
    """Formats the given value to a 4-digit hex string without leading "0x".

    Args:
        value (int | str | StrEnum): Value to format.

    Returns:
        str: Formatted value.
    """

    if isinstance(value, StrEnum):
        value = value.value
    if isinstance(value, int):
        value = f"0x{value:X}"  # Integer -> Hex string
    if value.startswith("0x"):  # Strip off leading 0x
        value = value[2:]
    return value.zfill(4)


def _create_write_cmd_string(
    cmd_address: CommandAddress | str | int,
    cmd_identifier: GeneralCommandIdentifier | TSMCommandIdentifier | M2MDCommandIdentifier | HKCommandIdentifier | str,
    cargo1: int = 0,
    cargo2: int = 0,
) -> str:
    """Creates a write-command string to send to the TCU Arduino.

    Note that this function should not be called directly in the `@dynamic_command` decorator (in the `cmd_string`
    attribute), as this will increase the transaction identifier without actually calling the function (merely
    importing the interface in which such a decorator call is made will increase the transaction identifier).

    Args:
        cmd_address (CommandAddress | str | int): Identifier of the commanded device.
        cmd_identifier (GeneralCommandIdentifier | TSMCommandIdentifier | M2MDCommandIdentifier | HKCommandIdentifier):
                       Command identifier, internal to the commanded device.
        cargo1 (int): Reference to the first 16-bit cargo word.
        cargo2 (int): Reference to the second 16-bit cargo word.

    Returns:
        Write-command string to send to the TCU Arduino.
    """

    return _create_cmd_string(PacketType.WRITE, cmd_address, cmd_identifier, cargo1, cargo2)


def _create_read_cmd_string(
    cmd_address: CommandAddress | str | int,
    cmd_identifier: GeneralCommandIdentifier | TSMCommandIdentifier | M2MDCommandIdentifier | HKCommandIdentifier | str,
    cargo1: int = 0,
    cargo2: int = 0,
) -> str:
    """Creates a read-command string to send to the TCU Arduino.

    Note that this function should not be called directly in the `@dynamic_command` decorator (in the `cmd_string`
    attribute), as this will increase the transaction identifier without actually calling the function (merely
    importing the interface in which such a decorator call is made will increase the transaction identifier)>

    Args:
        cmd_address (CommandAddress | str | int): Identifier of the commanded device.
        cmd_identifier (GeneralCommandIdentifier | TSMCommandIdentifier | M2MDCommandIdentifier | HKCommandIdentifier):
                       Command identifier, internal to the commanded device.
        cargo1 (int): Reference to the first 16-bit cargo word.
        cargo2 (int): Reference to the second 16-bit cargo word.

    Returns:
        Read-command string to send to the TCU Arduino.
    """

    return _create_cmd_string(PacketType.READ, cmd_address, cmd_identifier, cargo1, cargo2)


@static_vars(transaction_id=-1)
def _create_cmd_string(
    packet_type: PacketType,
    cmd_address: CommandAddress | str | int,
    cmd_identifier: GeneralCommandIdentifier | TSMCommandIdentifier | M2MDCommandIdentifier | HKCommandIdentifier | str,
    cargo1: int = 0,
    cargo2: int = 0,
) -> str:
    """Creates a command string to send to the TCU Arduino.

    Packet format (text):
        "03XX TTTT 0004 AAAA BBBB CCCC DDDD CRC"
    with:
        - 03: TCU logical address;
        - XX: Indicates whether it's a read (40) or write (20) command (without reply);
        - TTTT: Transaction identifier (basically a counter that increments for each command call);
        - 0004: Data length (always 4 bytes);
        - AAAA: Identifier of the commanded device:
            - 0000: General commands;
            - 0001: M2MD axis-1 commands;
            - 0002: M2MD axis-2 commands;
            - 0003: M2MD axis-3 commands;
            - 0004: TSM commands;
            - 0005: HK commands
        - BBBB: Command identifier, internal to the commanded device;
        - CCCC: First 16-bit cargo word;
        - DDDD: Second 16-bit cargo word;
        - CRC: Cyclic Redundancy Check (CRC-16), determined from the packet string.

    Note that this function should not be called directly in the `@dynamic_command` decorator (in the `cmd_string`
    attribute), as this will increase the transaction identifier without actually calling the function (merely
    importing the interface in which such a decorator call is made will increase the transaction identifier).

    Args:
        packet_type (PacketType): Type of the packet (read or write).
        cmd_address (CommandAddress): Identifier of the commanded device.  In case of a M2MD axis, it is the reference
                                      to the axis (typically "${axis}") rather than the axis identifier enumeration.
        cmd_identifier (GeneralCommandIdentifier | TSMCommandIdentifier | M2MDCommandIdentifier | HKCommandIdentifier
                       | str): Command identifier, internal to the commanded device.
        cargo1 (int): Reference to the first 16-bit cargo word.
        cargo2 (int): Reference to the second 16-bit cargo word.

    Returns:
        Command string to send to the TCU Arduino.
    """

    _create_cmd_string.transaction_id += 1

    # Adopted from Vladimiro's code

    address = format_value(cmd_address)
    cmd_identifier = format_value(cmd_identifier)
    cargo1 = format_value(cargo1)
    cargo2 = format_value(cargo2)

    transaction_id = f"0x{_create_cmd_string.transaction_id:X}"[2:].zfill(4)

    cmd_string = (
        f"{TCU_LOGICAL_ADDRESS}{packet_type.value} {transaction_id} {DATA_LENGTH} "
        f"{address} {cmd_identifier} {cargo1} {cargo2}"
    )

    cmd_string = append_crc16(cmd_string)
    return cmd_string


def create_crc16(cmd_str: str, ln: int = 14) -> str:
    """Calculates the 16-bit Cyclic Redundancy Check (CRC) checksum for the given command string.

    The CRC-16 is an error-detecting code that generates a 16-bit checksum to verify data integrity during
    transmission.  It works by performing a polynomial division (using XOR operations) on the data, with the remainder
    becoming the CRC checksum.  The receiver performs the same division: A zero remainder indicates the data arrived
    uncorrupted, while a non-zero remainder signals a potential error.

    The leading "0x" is stripped off and the CRC is padded with leading zeros to ensure it is always 4 characters long.

    Args:
        cmd_str (str): Command string for which the CRC will be calculated.
        ln (int): Number of bytes to include in the CRC calculation.

    Returns:
        Calculated 16-bit CRC checksum.
    """

    # Adapted from Vladimiro's code

    byte_array = bytearray([int(cmd_str[2 * i : 2 * i + 2], 16) for i in range(ln)])

    crc16_func = crcmod.mkCrcFun(0x11021, 0xFFFF, False, 0x0)
    crc16 = hex(crc16_func(byte_array))

    crc16 = crc16[2:6].zfill(4)  # Crop leading "0x" + ensure 4 characters

    return crc16


def append_crc16(cmd_string: str, ln: int = 14) -> str:
    """Appends a 16-bit CRC checksum to the given command string.

    Args:
        cmd_string (str): Command string to which the CRC will be appended.
        ln (int): Number of bytes to include in the CRC calculation.

    Returns:
        Command string with the appended CRC checksum.
    """

    crc16 = create_crc16(cmd_string, ln)

    return f"{cmd_string} {crc16}"


# General commands


def tcu_firmware_id() -> str:
    """Builds the command string for the general TCU_FIRMWARE_ID read command.

    Returns:
        Command string for the general TCU_FIRMWARE_ID read command.
    """

    return _create_read_cmd_string(CommandAddress.GENERAL, GeneralCommandIdentifier.TCU_FIRMWARE_ID)


def get_tcu_mode() -> str:
    """Builds the command string for the general TCU_MODE read command.

    Returns:
        Command string for the general TCU_MODE read command.
    """

    return _create_read_cmd_string(CommandAddress.GENERAL, GeneralCommandIdentifier.TCU_MODE)


def set_tcu_mode(tcu_mode: TcuMode | int = TcuMode.IDLE):
    """Builds the command string for the general TCU_MODE write command.

    Args:
        tcu_mode (TcuMode | int): TCU mode.

    Returns:
        Command string for the general TCU_MODE write command.
    """

    return _create_write_cmd_string(CommandAddress.GENERAL, GeneralCommandIdentifier.TCU_MODE, cargo2=tcu_mode)


def tcu_status() -> str:
    """Builds the command string for the general TCU_STATUS read command.

    Returns:
        Command string for the general TCU_STATUS read command.
    """

    return _create_read_cmd_string(CommandAddress.GENERAL, GeneralCommandIdentifier.TCU_STATUS)


def tcu_simulated(cargo2: int):
    """Builds the command string for the general TCU_SIMULATED write command.

    Args:
        cargo2 (int): Cargo 2 part of the command string.

    Returns:
        Command string for the general TCU_SIMULATED write command.
    """

    return _create_write_cmd_string(CommandAddress.GENERAL, GeneralCommandIdentifier.TCU_SIMULATED, cargo2=cargo2)


def restart_links_period_latch(cargo2: int):
    """Builds the command string for the general RESTART_LINKS_PERIOD_LATCH write command.

    Args:
        cargo2 (int): Cargo 2 part of the command string.

    Returns:
        Command string for the general RESTART_LINKS_PERIOD_LATCH write command.
    """

    return _create_write_cmd_string(
        CommandAddress.GENERAL, GeneralCommandIdentifier.RESTART_LINKS_PERIOD_LATCH, cargo2=cargo2
    )


def get_restart_links_period() -> str:
    """Builds the command string for the general RESTART_LINKS_PERIOD_LATCH read command.

    Returns:
        Command string for the general RESTART_LINKS_PERIOD_LATCH read command.
    """

    return _create_read_cmd_string(CommandAddress.GENERAL, GeneralCommandIdentifier.RESTART_LINKS_PERIOD)


def set_restart_links_period(link_period: int = 0xFFFF) -> str:
    """Builds the command string for the general RESTART_LINKS_PERIOD write command.

    Args:
        link_period (int): Link period.

    Returns:
        Command string for the general RESTART_LINKS_PERIOD write command.
    """

    return _create_write_cmd_string(
        CommandAddress.GENERAL,
        GeneralCommandIdentifier.RESTART_LINKS_PERIOD,
        cargo2=link_period,
    )


# M2MD commands


def ope_mng_command(axis: CommandAddress | str | int, cargo2: int = 0x0002) -> str:
    """Builds the command string fort he M2MD OPE_MNG_COMMAND write command.

    Args:
        axis (CommandAddress | str | int): Axis to which the command is sent.
        cargo2 (int): Cargo 2 part of the command string.

    Returns:
        Command string for the M2MD OPE_MNG_COMMAND write command.
    """

    return _create_write_cmd_string(axis, M2MDCommandIdentifier.OPE_MNG_COMMAND, cargo2=cargo2)


def ope_mng_event_clear_protect_flag(axis: CommandAddress | str | int, cargo2: str | int = 0xAAAA) -> str:
    """Builds the command string for the M2MD OPE_MNG_EVENT_CLEAR_PROTECT_FLAG write command.

    Args:
        axis (CommandAddress | str | int): Axis to which the command is sent.
        cargo2 (int): Cargo 2 part of the command string.

    Returns:
        Command string for the M2MD OPE_MNG_EVENT_CLEAR_PROTECT_FLAG write command.
    """

    return _create_write_cmd_string(axis, M2MDCommandIdentifier.OPE_MNG_EVENT_CLEAR_PROTECT_FLAG, cargo2=cargo2)


def ope_mng_event_clear(axis: CommandAddress | str | int, cargo2: int = 0) -> str:
    """Builds the command string for the M2MD OPE_MNG_EVENT_CLEAR write command.

    Args:
        axis (CommandAddress | str | int): Axis to which the command is sent.
        cargo2 (int): Cargo 2 part of the command string.

    Returns:
        Command string for the M2MD OPE_MNG_EVENT_CLEAR write command.
    """

    return _create_write_cmd_string(axis, M2MDCommandIdentifier.OPE_MNG_EVENT_CLEAR, cargo2=cargo2)


def ope_mng_status(axis: CommandAddress | str | int) -> str:
    """Builds the command string for the M2MD OPE_MNG_STATUS read command.

    Args:
        axis (CommandAddress | str | int): Axis to which the command is sent.

    Returns:
        Command string for the M2MD OPE_MNG_STATUS read command.
    """

    return _create_read_cmd_string(axis, M2MDCommandIdentifier.OPE_MNG_STATUS)


def ope_mng_event_reg(axis: CommandAddress | str | int) -> str:
    """Builds the command string for the M2MD OPE_MNG_EVENT_REG read command.

    Args:
        axis (CommandAddress | str | int): Axis to which the command is sent.

    Returns:
        Command string for the M2MD OPE_MNG_EVENT_REG read command.
    """

    return _create_read_cmd_string(axis, M2MDCommandIdentifier.OPE_MNG_EVENT_REG)


def get_acq_curr_off_corr(axis: CommandAddress | str | int) -> str:
    """Builds the command string for the M2MD ACQ_CURR_OFF_CORR read command.

    Args:
        axis (CommandAddress | str | int): Axis to which the command is sent.

    Returns:
        Command string for the M2MD ACQ_CURR_OFF_CORR read command.
    """

    return _create_read_cmd_string(axis, M2MDCommandIdentifier.ACQ_CURR_OFF_CORR)


def set_acq_curr_off_corr(axis: CommandAddress | str | int, cargo2: int = 0x03FB) -> str:
    """Builds the command string for the M2MD ACQ_CURR_OFF_CORR write command.

    Args:
        axis (CommandAddress | str | int): Axis to which the command is sent.
        cargo2 (int): Cargo 2 part of the command string.

    Returns:
        Command string for the M2MD ACQ_CURR_OFF_CORR write command.
    """

    return _create_write_cmd_string(axis, M2MDCommandIdentifier.ACQ_CURR_OFF_CORR, cargo2=cargo2)


def get_acq_curr_gain_corr(axis: CommandAddress | str | int) -> str:
    """Builds the command string for the M2MD ACQ_CURR_GAIN_CORR read command.

    Args:
        axis (CommandAddress | str | int): Axis to which the command is sent.

    Returns:
        Command string for the M2MD ACQ_CURR_GAIN_CORR read command.
    """

    return _create_read_cmd_string(axis, M2MDCommandIdentifier.ACQ_CURR_GAIN_CORR)


def set_acq_curr_gain_corr(axis: CommandAddress | str | int, cargo2: int = 0x074C):
    """Builds the command string for the M2MD ACQ_CURR_GAIN_CORR write command.

    Args:
        axis (CommandAddress | str | int): Axis to which the command is sent.
        cargo2 (int): Cargo 2 part of the command string.

    Returns:
        Command string for the M2MD ACQ_CURR_GAIN_CORR write command.
    """

    return _create_write_cmd_string(axis, M2MDCommandIdentifier.ACQ_CURR_GAIN_CORR, cargo2=cargo2)


def acq_axis_a_curr_read(axis: CommandAddress | str | int) -> str:
    """Builds the command string for the M2MD ACQ_AXIS_A_CURR_READ read command.

    Args:
        axis (CommandAddress | str | int): Axis to which the command is sent.

    Returns:
        Command string for the M2MD ACQ_AXIS_A_CURR_READ read command.
    """

    return _create_read_cmd_string(axis, M2MDCommandIdentifier.ACQ_AXIS_A_CURR_READ)


def acq_axis_b_curr_read(axis: CommandAddress | str | int) -> str:
    """Builds the command string for the M2MD ACQ_AXIS_B_CURR_READ read command.

    Args:
        axis (CommandAddress | str | int): Axis to which the command is sent.

    Returns:
        Command string for the M2MD ACQ_AXIS_B_CURR_READ read command.
    """

    return _create_read_cmd_string(axis, M2MDCommandIdentifier.ACQ_AXIS_B_CURR_READ)


def acq_ave_lpf_en(axis: CommandAddress | str | int, cargo2: int = 0) -> str:
    """Builds the command string for the M2MD ACQ_AVE_LPF_EN write command.

    Args:
        axis (CommandAddress | str | int): Axis to which the command is sent.
        cargo2 (int): Cargo 2 part of the command string.

    Returns:
        Command string for the M2MD ACQ_AVE_LPF_EN write command.
    """

    return _create_write_cmd_string(axis, M2MDCommandIdentifier.ACQ_AVE_LPF_EN, cargo2=cargo2)


def acq_ovc_cfg_filter(axis: CommandAddress | str | int, cargo2: int = 0) -> str:
    """Builds the command string for the M2MD ACQ_OVC_CFG_FILTER write command.

    Args:
        axis (CommandAddress | str | int): Axis to which the command is sent.
        cargo2 (int): Cargo 2 part of the command string.

    Returns:
        Command string for the M2MD ACQ_OVC_CFG_FILTER write command.
    """

    return _create_write_cmd_string(axis, M2MDCommandIdentifier.ACQ_OVC_CFG_FILTER, cargo2=cargo2)


def acq_avc_filt_time(axis: CommandAddress | str | int, cargo2: int = 0) -> str:
    """Builds the command string for the M2MD ACQ_AVC_FILT_TIME write command.

    Args:
        axis (CommandAddress | str | int): Axis to which the command is sent.
        cargo2 (int): Cargo 2 part of the command string.

    Returns:
        Command string for the M2MD ACQ_AVC_FILT_TIME write command.
    """

    return _create_write_cmd_string(axis, M2MDCommandIdentifier.ACQ_AVC_FILT_TIME, cargo2=cargo2)


def acq_average_type(axis: CommandAddress | str | int, cargo2: int = 0x0000) -> str:
    """Builds the command string for the M2MD ACQ_AVERAGE_TYPE write command.

    Args:
        axis (CommandAddress | str | int): Axis to which the command is sent.
        cargo2 (int): Cargo 2 part of the command string.

    Returns:
        Command string for the M2MD ACQ_AVERAGE_TYPE write command.
    """

    return _create_write_cmd_string(axis, M2MDCommandIdentifier.ACQ_AVERAGE_TYPE, cargo2=cargo2)


def acq_spk_filt_counter_lim(axis: CommandAddress | str | int, cargo2: int = 0x0001) -> str:
    """Builds the command string for the M2MD ACQ_SPK_FILT_COUNTER_LIM write command.

    Args:
        axis (CommandAddress | str | int): Axis to which the command is sent.
        cargo2 (int): Cargo 2 part of the command string.

    Returns:
        Command string for the M2MD ACQ_SPK_FILT_COUNTER_LIM write command.
    """

    return _create_write_cmd_string(axis, M2MDCommandIdentifier.ACQ_SPK_FILT_COUNTER_LIM, cargo2=cargo2)


def acq_spk_filt_incr_thr(axis: CommandAddress | str | int, cargo2: int = 0x04C0) -> str:
    """Builds the command string for the M2MD ACQ_SPK_FILT_INCR_THR write command.

    Args:
        axis (CommandAddress | str | int): Axis to which the command is sent.
        cargo2 (int): Cargo 2 part of the command string.

    Returns:
        Command string for the M2MD ACQ_SPK_FILT_INCR_THR write command.
    """

    return _create_write_cmd_string(axis, M2MDCommandIdentifier.ACQ_SPK_FILT_INCR_THR, cargo2=cargo2)


def get_prof_gen_axis_step(axis: CommandAddress | str | int) -> str:
    """Builds the command string for the M2MD PROF_GEN_AXIS_STEP read command.

    Args:
        axis (CommandAddress | str | int): Axis to which the command is sent.

    Returns:
        Command string for the M2MD PROF_GEN_AXIS_STEP read command.
    """

    return _create_read_cmd_string(axis, M2MDCommandIdentifier.PROF_GEN_AXIS_STEP)


def set_prof_gen_axis_step(axis: CommandAddress | str | int, cargo2: int = 0x0480) -> str:
    """Builds the command string for the M2MD PROF_GEN_AXIS_STEP write command.

    Args:
        axis (CommandAddress | str | int): Axis to which the command is sent.
        cargo2 (int): Cargo 2 part of the command string.

    Returns:
        Command string for the M2MD PROF_GEN_AXIS_STEP write command.
    """

    return _create_write_cmd_string(axis, M2MDCommandIdentifier.PROF_GEN_AXIS_STEP, cargo2=cargo2)


def get_prof_gen_axis_speed(axis: CommandAddress | str | int) -> str:
    """Builds the command string for the M2MD PROF_GEN_AXIS_SPEED read command.

    Args:
        axis (CommandAddress | str | int): Axis to which the command is sent.

    Returns:
        Command string for the M2MD PROF_GEN_AXIS_SPEED read command.
    """

    return _create_read_cmd_string(axis, M2MDCommandIdentifier.PROF_GEN_AXIS_SPEED)


def set_prof_gen_axis_speed(axis: CommandAddress | str | int, speed: int = 0x1777) -> str:
    """Builds the command string for the M2MD PROF_GEN_AXIS_SPEED write command.

    Args:
        axis (CommandAddress | str | int): Axis to which the command is sent.
        speed (int): Cargo 2 part of the command string.

    Returns:
        Command string for the M2MD PROF_GEN_AXIS_SPEED write command.
    """

    return _create_write_cmd_string(axis, M2MDCommandIdentifier.PROF_GEN_AXIS_SPEED, cargo2=speed)


def get_prof_gen_axis_state_start(axis: CommandAddress | str | int) -> str:
    """Builds the command string for the M2MD PROF_GEN_AXIS_STATE_START read command.

    Args:
        axis (CommandAddress | str | int): Axis to which the command is sent.

    Returns:
        Command string for the M2MD PROF_GEN_AXIS_STATE_START read command.
    """

    return _create_read_cmd_string(axis, M2MDCommandIdentifier.PROF_GEN_AXIS_STATE_START)


def set_prof_gen_axis_state_start(axis: CommandAddress | str | int, cargo2: int = 0) -> str:
    """Builds the command string for the M2MD PROF_GEN_AXIS_STATE_START write command.

    Args:
        axis (CommandAddress | str | int): Axis to which the command is sent.
        cargo2 (int): Cargo 2 part of the command string.

    Returns:
        Command string for the M2MD PROF_GEN_AXIS_STATE_START write command.
    """

    return _create_write_cmd_string(axis, M2MDCommandIdentifier.PROF_GEN_AXIS_STATE_START, cargo2=cargo2)


def sw_rs_xx_sw_rise(axis: CommandAddress | str | int, position: int = 1) -> str:
    """Builds the command string for the M2MD SW_RS_XX_SW_RISE read command.

    Args:
        axis (CommandAddress | str | int): Axis to which the command is sent.
        position (int): Position of the SW_RS_XX_SW_RISE command.

    Returns:
        Command string for the M2MD SW_RS_XX_SW_RISE read command.
    """

    # Adopted from Vladimiro's code
    m2md_cmd_id = get_sw_rs_xx_sw_cmd_id(M2MDCommandIdentifier.SW_RS_XX_SW_RISE, position)

    return _create_read_cmd_string(axis, cmd_identifier=m2md_cmd_id)


def sw_rs_xx_sw_fall(axis: CommandAddress | str | int, position: int = 1) -> str:
    """Builds the command string for the M2MD SW_RS_XX_SW_FALL read command.

    Args:
        axis (CommandAddress | str | int): Axis to which the command is sent.
        position (int): Position of the SW_RS_XX_SW_FALL command.

    Returns:
        Command string for the M2MD SW_RS_XX_SW_FALL read command.
    """

    # Adopted from Vladimiro's code
    offset_position = position + 21
    m2md_cmd_id = get_sw_rs_xx_sw_cmd_id(M2MDCommandIdentifier.SW_RS_XX_SW_FALL, offset_position)

    return _create_read_cmd_string(axis, cmd_identifier=m2md_cmd_id)


# TSM commands


def tsm_latch(cargo1: int = 0, cargo2: int = 0) -> str:
    """Builds the command string for the TSM TSM_LATCH write command.

    Args:
        cargo1 (int): Cargo 1 part of the command string.
        cargo2 (int): Cargo 2 part of the command string.

    Returns:
        Command string for the TSM TSM_LATCH write command.
    """

    return _create_write_cmd_string(CommandAddress.TSM, TSMCommandIdentifier.TSM_LATCH, cargo1=cargo1, cargo2=cargo2)


def get_tsm_current_value() -> str:
    """Builds the command string for the TSM TSM_CURRENT_VALUE read command.

    Returns:
        Command string for the TSM TSM_CURRENT_VALUE read command.
    """

    return _create_read_cmd_string(CommandAddress.TSM, TSMCommandIdentifier.TSM_CURRENT_VALUE)


def set_tsm_current_value(cargo1: int = 0, cargo2: int = 0) -> str:
    """Builds the command string for the TSM TSM_CURRENT_VALUE write command.

    Args:
        cargo1 (int): Cargo 1 part of the command string.
        cargo2 (int): Cargo 2 part of the command string.

    Returns:
        Command string for the TSM TSM_CURRENT_VALUE write command.
    """

    return _create_write_cmd_string(
        CommandAddress.TSM, TSMCommandIdentifier.TSM_CURRENT_VALUE, cargo1=cargo1, cargo2=cargo2
    )


def get_tsm_current_offset() -> str:
    """Builds the command string for the TSM TSM_CURRENT_OFFSET read command.

    Returns:
        Command string for the TSM TSM_CURRENT_OFFSET read command.
    """

    return _create_read_cmd_string(CommandAddress.TSM, TSMCommandIdentifier.TSM_CURRENT_OFFSET)


def set_tsm_current_offset(cargo1: int = 0, cargo2: int = 0) -> str:
    """Builds the command string for the TSM TSM_CURRENT_OFFSET write command.

    Args:
        cargo1 (int): Cargo 1 part of the command string.
        cargo2 (int): Cargo 2 part of the command string.

    Returns:
        Command string for the TSM TSM_CURRENT_OFFSET write command.
    """

    return _create_write_cmd_string(
        CommandAddress.TSM, TSMCommandIdentifier.TSM_CURRENT_OFFSET, cargo1=cargo1, cargo2=cargo2
    )


def tsm_adc_register_latch(cargo1: int = 0, cargo2: int = 0) -> str:
    """Builds the command string for the TSM TSM_ADC_REGISTER_LATCH write command.

    Args:
        cargo1 (int): Cargo 1 part of the command string.
        cargo2 (int): Cargo 2 part of the command string.

    Returns:
        Command string for the TSM TSM_ADC_REGISTER_LATCH write command.
    """

    return _create_write_cmd_string(
        CommandAddress.TSM, TSMCommandIdentifier.TSM_ADC_REGISTER_LATCH, cargo1=cargo1, cargo2=cargo2
    )


def tsm_adc_id_register() -> str:
    """Builds the command string for the TSM TSM_ADC_ID_REGISTER read command.

    Returns:
        Command string for the TSM TSM_ADC_ID_REGISTER read command.
    """

    return _create_read_cmd_string(CommandAddress.TSM, TSMCommandIdentifier.TSM_ADC_ID_REGISTER)


def tsm_adc_configuration_register() -> str:
    """Builds the command string for the TSM TSM_ADC_CONFIGURATION_REGISTER read command.

    Returns:
        Command string for the TSM TSM_ADC_CONFIGURATION_REGISTER read command.
    """

    return _create_read_cmd_string(CommandAddress.TSM, TSMCommandIdentifier.TSM_ADC_CONFIGURATION_REGISTER)


def get_tsm_adc_hpf_register() -> str:
    """Builds the command string for the TSM TSM_ADC_HPF_REGISTER read command.

    Returns:
        Command string for the TSM TSM_ADC_HPF_REGISTER read command.
    """

    return _create_read_cmd_string(CommandAddress.TSM, TSMCommandIdentifier.TSM_ADC_HPF_REGISTER)


def set_tsm_adc_hpf_register(cargo1: int = 0, cargo2: int = 0) -> str:
    """Builds the command string for the TSM TSM_ADC_HPF_REGISTER write command.

    Args:
        cargo1 (int): Cargo 1 part of the command string.
        cargo2 (int): Cargo 2 part of the command string.

    Returns:
        Command string for the TSM TSM_ADC_HPF_REGISTER write command.
    """

    return _create_write_cmd_string(
        CommandAddress.TSM, TSMCommandIdentifier.TSM_ADC_HPF_REGISTER, cargo1=cargo1, cargo2=cargo2
    )


def get_tsm_adc_ofc_register() -> str:
    """Builds the command string for the TSM TSM_ADC_OFC_REGISTER read command.

    Returns:
        Command string for the TSM TSM_ADC_OFC_REGISTER read command.
    """

    return _create_read_cmd_string(CommandAddress.TSM, TSMCommandIdentifier.TSM_ADC_OFC_REGISTER)


def set_tsm_adc_ofc_register(cargo1: int = 0, cargo2: int = 0) -> str:
    """Builds the command string for the TSM TSM_ADC_OFC_REGISTER write command.

    Args:
        cargo1 (int): Cargo 1 part of the command string.
        cargo2 (int): Cargo 2 part of the command string.

    Returns:
        Command string for the TSM TSM_ADC_OFC_REGISTER write command.
    """

    return _create_write_cmd_string(
        CommandAddress.TSM, TSMCommandIdentifier.TSM_ADC_OFC_REGISTER, cargo1=cargo1, cargo2=cargo2
    )


def get_tsm_adc_fsc_register() -> str:
    """Builds the command string for the TSM TSM_ADC_FSC_REGISTER read command.

    Returns:
        Command string for the TSM TSM_ADC_FSC_REGISTER read command.
    """

    return _create_read_cmd_string(CommandAddress.TSM, TSMCommandIdentifier.TSM_ADC_FSC_REGISTER)


def set_tsm_adc_fsc_register(cargo1: int = 0, cargo2: int = 0) -> str:
    """Builds the command string for the TSM TSM_ADC_FSC_REGISTER write command.

    Args:
        cargo1 (int): Cargo 1 part of the command string.
        cargo2 (int): Cargo 2 part of the command string.

    Returns:
        Command string for the TSM TSM_ADC_FSC_REGISTER write command.
    """

    return _create_write_cmd_string(
        CommandAddress.TSM, TSMCommandIdentifier.TSM_ADC_FSC_REGISTER, cargo1=cargo1, cargo2=cargo2
    )


def tsm_adc_command_latch(cargo1: int = 0, cargo2: int = 0) -> str:
    """Builds the command string for the TSM TSM_ADC_COMMAND_LATCH write command.

    Args:
        cargo1 (int): Cargo 1 part of the command string.
        cargo2 (int): Cargo 2 part of the command string.

    Returns:
        Command string for the TSM TSM_ADC_COMMAND_LATCH write command.
    """

    return _create_write_cmd_string(
        CommandAddress.TSM, TSMCommandIdentifier.TSM_ADC_COMMAND_LATCH, cargo1=cargo1, cargo2=cargo2
    )


def tsm_adc_command(cargo1: int = 0, cargo2: int = 0) -> str:
    """Builds the command string for the TSM TSM_ADC_COMMAND write command.

    Args:
        cargo1 (int): Cargo 1 part of the command string.
        cargo2 (int): Cargo 2 part of the command string.

    Returns:
        Command string for the TSM TSM_ADC_COMMAND write command.
    """

    return _create_write_cmd_string(
        CommandAddress.TSM, TSMCommandIdentifier.TSM_ADC_COMMAND, cargo1=cargo1, cargo2=cargo2
    )


def tsm_adc_calibration(cargo1: int = 0, cargo2: int = 0) -> str:
    """Builds the command string for the TSM TSM_ADC_CALIBRATION write command.

    Args:
        cargo1 (int): Cargo 1 part of the command string.
        cargo2 (int): Cargo 2 part of the command string.

    Returns:
        Command string for the TSM TSM_ADC_CALIBRATION write command.
    """

    return _create_write_cmd_string(
        CommandAddress.TSM, TSMCommandIdentifier.TSM_ADC_CALIBRATION, cargo1=cargo1, cargo2=cargo2
    )


def tsm_adc_value_xx_currentn(probe: int = 1) -> str:
    """Builds the command string for the TSM TSM_ADC_VALUE_XX_CURRENTN read command.

    Args:
        probe (int): Thermistor identifier.

    Returns:
        Command string for the TSM TSM_ADC_VALUE_XX_CURRENTN read command.
    """

    tsm_cmd_id = get_tsm_adc_value_xx_currentn_cmd_id(probe)

    return _create_read_cmd_string(CommandAddress.TSM, tsm_cmd_id)


def tsm_adc_value_xx_biasn(probe: int = 1):
    """Builds the command string for the TSM TSM_ADC_VALUE_XX_BIASN read command.

    Args:
        probe (int): Thermistor identifier.

    Returns:
        Command string for the TSM TSM_ADC_VALUE_XX_BIASN read command.
    """

    tsm_cmd_id = get_tsm_adc_value_xx_biasn_cmd_id(probe)

    return _create_read_cmd_string(CommandAddress.TSM, tsm_cmd_id)


def tsm_adc_value_xx_currentp(probe: int = 1) -> str:
    """Builds the command string for the TSM TSM_ADC_VALUE_XX_CURRENTP read command.

    Args:
        probe (int): Thermistor identifier.

    Returns:
        Command string for the TSM TSM_ADC_VALUE_XX_CURRENTP read command.
    """

    tsm_cmd_id = get_tsm_adc_value_xx_currentp_cmd_id(probe)

    return _create_read_cmd_string(CommandAddress.TSM, tsm_cmd_id)


def tsm_adc_value_xx_biasp(probe: int = 1) -> str:
    """Builds the command string for the TSM TSM_ADC_VALUE_XX_BIASP read command.

    Args:
        probe (int): Thermistor identifier.

    Returns:
        Command string for the TSM TSM_ADC_VALUE_XX_BIASP read command.
    """

    tsm_cmd_id = get_tsm_adc_value_xx_biasp_cmd_id(probe)

    return _create_read_cmd_string(CommandAddress.TSM, tsm_cmd_id)


def tsm_acq_counter() -> str:
    """Builds the command string for the TSM TSM_ACQ_COUNTER read command.

    Returns:
        Command string for the TSM TSM_ACQ_COUNTER read command.
    """

    return _create_read_cmd_string(CommandAddress.TSM, TSMCommandIdentifier.TSM_ACQ_COUNTER)


def vhk_psu_vmotor() -> str:
    """Builds the command string for the HK VHK_PSU_VMOTOR read command.

    Returns:
        Command string for the HK PSU VHK_PSU_VMOTOR read command.
    """

    return _create_read_cmd_string(CommandAddress.HK, HKCommandIdentifier.VHK_PSU_VMOTOR)


def vhk_psu_vhi() -> str:
    """Builds the command string for the HK VHK_PSU_VHI read command.

    Returns:
        Command string for the HK PSU VHK_PSU_VHI read command.
    """

    return _create_read_cmd_string(CommandAddress.HK, HKCommandIdentifier.VHK_PSU_VHI)


def vhk_psu_vlow() -> str:
    """Builds the command string for the HK VHK_PSU_VLW read command.

    Returns:
        Command string for the HK PSU VHK_PSU_VLOW read command.
    """

    return _create_read_cmd_string(CommandAddress.HK, HKCommandIdentifier.VHK_PSU_VLOW)


def vhk_psu_vmedp() -> str:
    """Builds the command string for the HK VHK_PSU_VMEDP read command.

    Returns:
        Command string for the HK PSU VHK_PSU_VMEDP read command.
    """

    return _create_read_cmd_string(CommandAddress.HK, HKCommandIdentifier.VHK_PSU_VMEDP)


def vhk_psu_vmedn() -> str:
    """Builds the command string for the HK VHK_PSU_VMEDN read command.

    Returns:
        Command string for the HK PSU VHK_PSU_VMEDN read command.
    """

    return _create_read_cmd_string(CommandAddress.HK, HKCommandIdentifier.VHK_PSU_VMEDN)


def ihk_psu_vmedn() -> str:
    """Builds the command string for the HK IHK_PSU_VMEDN read command.

    Returns:
        Command string for the HK PSU IHK_PSU_VMEDN read command.
    """

    return _create_read_cmd_string(CommandAddress.HK, HKCommandIdentifier.IHK_PSU_VMEDN)


def ihk_psu_vmedp() -> str:
    """Builds the command string for the HK IHK_PSU_VMEDP read command.

    Returns:
        Command string for the HK PSU IHK_PSU_VMEDP read command.
    """

    return _create_read_cmd_string(CommandAddress.HK, HKCommandIdentifier.IHK_PSU_VMEDP)


def ihk_psu_vlow() -> str:
    """Builds the command string for the HK IHK_PSU_VLOW read command.

    Returns:
        Command string for the HK PSU IHK_PSU_VLOW read command.
    """

    return _create_read_cmd_string(CommandAddress.HK, HKCommandIdentifier.IHK_PSU_VLOW)


def ihk_psu_vhi():
    """Builds the command string for the HK IHK_PSU_VHI read command.

    Returns:
        Command string for the HK PSU IHK_PSU_VHI read command.
    """

    return _create_read_cmd_string(CommandAddress.HK, HKCommandIdentifier.IHK_PSU_VHI)


def ihk_psu_vmotor() -> str:
    """Builds the command string for the HK IHK_PSU_VMOTOR read command.

    Returns:
        Command string for the HK PSU IHK_PSU_VMOTOR read command.
    """

    return _create_read_cmd_string(CommandAddress.HK, HKCommandIdentifier.IHK_PSU_VMOTOR)


def thk_psu_first() -> str:
    """Builds the command string for the HK THK_PSU_FIRST read command.

    Returns:
        Command string for the HK PSU THK_PSU_FIRST read command.
    """

    return _create_read_cmd_string(CommandAddress.HK, HKCommandIdentifier.THK_PSU_FIRST)


def thk_m2md_first() -> str:
    """Builds the command string for the HK THK_M2MD_FIRST read command.

    Returns:
        Command string for the HK M2MD THK_M2MD_FIRST read command.
    """

    return _create_read_cmd_string(CommandAddress.HK, HKCommandIdentifier.THK_M2MD_FIRST)


def thk_psu_second() -> str:
    """Builds the command string for the HK THK_PSU_SECOND read command.

    Returns:
        Command string for the HK PSU THK_PSU_SECOND read command.
    """

    return _create_read_cmd_string(CommandAddress.HK, HKCommandIdentifier.THK_PSU_SECOND)


def thk_m2md_second() -> str:
    """Builds the command string for the HK THK_M2MD_SECOND read command.

    Returns:
        Command string for the HK M2MD THK_M2MD_SECOND read command.
    """

    return _create_read_cmd_string(CommandAddress.HK, HKCommandIdentifier.THK_M2MD_SECOND)


def thk_cts_q1() -> str:
    """Builds the command string for the HK THK_CTS_Q1 read command.

    Returns:
        Command string for the HK CTS THK_CTS_Q1 read command.
    """

    return _create_read_cmd_string(CommandAddress.HK, HKCommandIdentifier.THK_CTS_Q1)


def thk_cts_q2() -> str:
    """Builds the command string for the HK THK_CTS_Q2 read command.

    Returns:
        Command string for the HK CTS THK_CTS_Q2 read command.
    """

    return _create_read_cmd_string(CommandAddress.HK, HKCommandIdentifier.THK_CTS_Q2)


def thk_cts_q3() -> str:
    """Builds the command string for the HK THK_CTS_Q3 read command.

    Returns:
        Command string for the HK CTS THK_CTS_Q3 read command.
    """

    return _create_read_cmd_string(CommandAddress.HK, HKCommandIdentifier.THK_CTS_Q3)


def thk_cts_q4() -> str:
    """Builds the command string for the HK THK_CTS_Q4 read command.

    Returns:
        Command string for the HK CTS THK_CTS_Q4 read command.
    """

    return _create_read_cmd_string(CommandAddress.HK, HKCommandIdentifier.THK_CTS_Q4)


def thk_cts_fpga() -> str:
    """Builds the command string for the HK THK_CTS_FPGA read command.

    Returns:
        Command string for the HK CTS THK_CTS_FPGA read command.
    """

    return _create_read_cmd_string(CommandAddress.HK, HKCommandIdentifier.THK_CTS_FPGA)


def thk_cts_ads1282() -> str:
    """Builds the command string for the HK THK_CTS_ADS1282 read command.

    Returns:
        Command string for the HK CTS THK_CTS_ADS1282 read command.
    """

    return _create_read_cmd_string(CommandAddress.HK, HKCommandIdentifier.THK_CTS_ADS1282)


def vhk_ths_ret() -> str:
    """Builds the command string for the HK VHK_THS_RET read command.

    Returns:
        Command string for the HK VHK_THS_RET read command.
    """

    return _create_read_cmd_string(CommandAddress.HK, HKCommandIdentifier.VHK_THS_RET)


def hk_acq_counter() -> str:
    """Builds the command string for the HK HK_ACQ_COUNTER read command.

    Returns:
        Command string for the HK HK_ACQ_COUNTER read command.
    """

    return _create_read_cmd_string(CommandAddress.HK, HKCommandIdentifier.HK_ACQ_COUNTER)


def general_cmd():
    """Decorator for general TCU commands.

    This is used to decorate these commands in the TcuInterface class, to enable an easier identification of the
    component of the TCU that the command is for (which is necessary in the TCU UI).
    """

    def decorator(func):
        func.target_comp = "GENERAL"
        return func

    return decorator


def m2md_cmd():
    """Decorator for M2MD commands.

    This is used to decorate these commands in the TcuInterface class, to enable an easier identification of the
    component of the TCU that the command is for (which is necessary in the TCU UI).
    """

    def decorator(func):
        func.target_comp = "M2MD"
        return func

    return decorator


def tsm_cmd():
    """Decorator for TSM commands.

    This is used to decorate these commands in the TcuInterface class, to enable an easier identification of the
    component of the TCU that the command is for (which is necessary in the TCU UI).
    """

    def decorator(func):
        func.target_comp = "TSM"
        return func

    return decorator


def hk_cmd():
    """Decorator for HK commands.

    This is used to decorate these commands in the TcuInterface class, to enable an easier identification of the
    component of the TCU that the command is for (which is necessary in the TCU UI).
    """

    def decorator(func):
        func.target_comp = "HK"
        return func

    return decorator



from glob import glob

import rich

# Quick hack to load plugins from different places.
# This doesn't work, I get an import error.

# def refactor(string: str) -> str:
#     entry = string.replace("/", ".").replace("\\", ".").replace(".py", "")
#     rich.print(f"[green]{entry = }[/]")
#     return entry
#
# pytest_plugins = [
#     refactor(conftest)
#     for conftest in glob("**/tests/fixtures/*.py", recursive=True)
#     if "__" not in conftest
# ]

# Doesn't work, with the following error:
#
# File "/Users/rik/github/cgse/libs/cgse-common/tests/fixtures/default_env.py", line 7, in <module>
#     from fixtures.helpers import setup_data_storage_layout, teardown_data_storage_layout
# ImportError: Error importing plugin "libs.cgse-common.tests.fixtures.default_env": No module named 'fixtures'
#

# pytest_plugins = [
#     "libs.cgse-common.tests.fixtures.helpers",
#     "libs.cgse-common.tests.fixtures.default_env",
#     "libs.cgse-core.tests.fixtures.services",
# ]



from egse.log import logger
from egse.notifyhub.services import EventSubscriber

subscriber = EventSubscriber([""])
subscriber.connect()


def handle_simple_event(event_data: dict):
    logger.info(f"----> {event_data=}")


def handle_single_event(event_data: dict):
    logger.info(f"----> {event_data=}")


subscriber.register_handler("simple-event", handle_simple_event)
subscriber.register_handler("single-event", handle_single_event)

while True:
    try:
        if subscriber.poll():
            logger.info("Event spotted...")
            event_data = subscriber.handle_event(return_event_data=True)
            logger.info(f"{event_data=}")
    except KeyboardInterrupt:
        logger.info("Caught KeyboardInterrupt, terminating...")
        break

subscriber.disconnect()



"""
A Graphical User Interface for monitoring and commanding the Symétrie Hexapod.

Start the GUI from your terminal as follows:

    puna_ui [--type proxy|direct|simulator]

This GUI is based on the SYM_positioning application from Symétrie. The intent
is to provide operators a user interface which is platform independent, but
familiar.

The application is completely written in Python/Qt5 and can therefore run on any
platform that supports Python and Qt5.

"""

import argparse
import multiprocessing
import sys
import threading
from enum import IntEnum
from pathlib import Path

import typer
from PyQt5.QtCore import QLockFile
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QApplication
from PyQt5.QtWidgets import QFrame
from PyQt5.QtWidgets import QHBoxLayout
from PyQt5.QtWidgets import QMessageBox
from PyQt5.QtWidgets import QVBoxLayout
from dotenv import find_dotenv
from dotenv import load_dotenv

from egse.gui import show_warning_message
from egse.gui.led import Indic
from egse.gui.states import States
from egse.hexapod.symetrie import ControllerFactory
from egse.hexapod.symetrie import ProxyFactory
from egse.hexapod.symetrie import get_hexapod_controller_pars
from egse.hexapod.symetrie.hexapod_ui import ActuatorStates
from egse.hexapod.symetrie.hexapod_ui import HexapodUIController
from egse.hexapod.symetrie.hexapod_ui import HexapodUIModel
from egse.hexapod.symetrie.hexapod_ui import HexapodUIView
from egse.hexapod.symetrie.puna import PunaSimulator
from egse.hexapod.symetrie.punaplus import PunaPlusController
from egse.hexapod.symetrie.punaplus import PunaPlusProxy
from egse.log import logging
from egse.process import ProcessStatus
from egse.resource import get_resource
from egse.settings import Settings
from egse.system import do_every

MODULE_LOGGER = logging.getLogger(__name__)

load_dotenv(find_dotenv(usecwd=True), override=True)


class DeviceControllerType(IntEnum):
    ALPHA = 0
    ALPHA_PLUS = 1


DCT = DeviceControllerType


# Status LEDs define the number of status leds (length of the list), the description and the
# default color when the LED is on.

STATUS_LEDS_ALPHA = [
    ["Error", Indic.RED],  # bit 0
    ["System Initialized", Indic.GREEN],  # bit 1
    ["In position", Indic.GREEN],  # bit 2
    ["Amplifier enabled", Indic.GREEN],  # bit 3
    ["Homing done", Indic.GREEN],  # bit 4
    ["Brake on", Indic.GREEN],  # bit 5
    ["Emergency stop", Indic.ORANGE],  # bit 6
    ["Warning FE", Indic.ORANGE],  # bit 7
    ["Fatal FE", Indic.RED],  # bit 8
    ["Actuator Limit Error", Indic.RED],  # bit 9
    ["Amplifier Error", Indic.RED],  # bit 10
    ["Encoder error", Indic.RED],  # bit 11
    ["Phasing error", Indic.RED],  # bit 12
    ["Homing error", Indic.RED],  # bit 13
    ["Kinematic error", Indic.RED],  # bit 14
    ["Abort input error", Indic.RED],  # bit 15
    ["R/W memory error", Indic.RED],  # bit 16
    ["Temperature error", Indic.RED],  # bit 17
    ["Homing done (virtual)", Indic.ORANGE],  # bit 18
    ["Encoders power off", Indic.ORANGE],  # bit 19
    ["Limit switches power off", Indic.ORANGE],  # bit 20
    ["Reserved", Indic.BLACK],  # bit 21
    ["Reserved", Indic.BLACK],  # bit 22
    ["Reserved", Indic.BLACK],  # bit 23
]

STATUS_LEDS_ALPHA_PLUS = [
    ["Error", Indic.RED],  # bit 0
    ["System Initialized", Indic.GREEN],  # bit 1
    ["Control On", Indic.GREEN],  # bit 2
    ["In Position", Indic.GREEN],  # bit 3
    ["Motion Task Running", Indic.GREEN],  # bit 4
    ["Home Task Running", Indic.GREEN],  # bit 5
    ["Home Complete", Indic.GREEN],  # bit 6
    ["Home Virtual", Indic.GREEN],  # bit 7
    ["Phase Found", Indic.GREEN],  # bit 8
    ["Brake on", Indic.GREEN],  # bit 9
    ["Motion Restricted", Indic.RED],  # bit 10
    ["Power on Encoders", Indic.GREEN],  # bit 11
    ["Power on Limit switches", Indic.GREEN],  # bit 12
    ["Power on Drives", Indic.GREEN],  # bit 13
    ["Emergency Stop", Indic.RED],  # bit 14
]

STATUS_LEDS = {DCT.ALPHA: STATUS_LEDS_ALPHA, DCT.ALPHA_PLUS: STATUS_LEDS_ALPHA_PLUS}

# The index of the Control LED

CONTROL_ONOFF = {DCT.ALPHA: 3, DCT.ALPHA_PLUS: 2}

ACTUATOR_STATE_LABELS_ALPHA = [
    "In position",
    "Control loop on servo motors active",
    "Homing done",
    "Input “Home switch”",
    "Input “Positive limit switch”",
    "Input “Negative limit switch”",
    "Brake control output",
    "Following error (warning)",
    "Following error",
    "Actuator out of bounds error",
    "Amplifier error",
    "Encoder error",
    "Phasing error (brushless engine only)",
]

ACTUATOR_STATE_LABELS_ALPHA_PLUS = [
    "Error: ",
    "Control On: ",
    "In Position: ",
    "Motion Task Running: ",
    "Home task running: ",
    "Home complete: ",
    "Phase found: ",
    "Brake on: ",
    "Home HW input: ",
    "Negative HW limit switch: ",
    "Positive HW limit switch: ",
    "SW limit reached: ",
    "Following Error: ",
    "Drive fault: ",
    "Encoder error: ",
]

ACTUATOR_STATE_LABELS = {DCT.ALPHA: ACTUATOR_STATE_LABELS_ALPHA, DCT.ALPHA_PLUS: ACTUATOR_STATE_LABELS_ALPHA_PLUS}


class PunaUIView(HexapodUIView):
    def __init__(self, device_controller_type: DCT, device_id: str):
        super().__init__()

        self.dct = device_controller_type

        if self.dct == DCT.ALPHA:
            title = f"Hexapod PUNA Controller (Alpha) – {device_id}"
        else:
            title = f"Hexapod PUNA Controller (Alpha+) – {device_id}"

        self.setWindowTitle(title)
        self.actuator_states = ActuatorStates(labels=ACTUATOR_STATE_LABELS[self.dct])

        self.init_gui()

    def init_gui(self):
        # The main frame in which all the other frames are located, the outer Application frame

        app_frame = QFrame()
        app_frame.setObjectName("AppFrame")

        # The left part which shows the states and positions

        status_frame = QFrame()
        status_frame.setObjectName("StatusFrame")

        # The right part which has tabs that allow settings, movements, maintenance etc.

        tabs_frame = QFrame()
        tabs_frame.setObjectName("TabsFrame")

        # The states of the Hexapod (contains all the leds)

        states_frame = QFrame()
        states_frame.setObjectName("StatesFrame")

        # The user, machine positions and actuator lengths

        positions_frame = QFrame()
        positions_frame.setObjectName("PositionsFrame")

        hbox = QHBoxLayout()
        vbox_left = QVBoxLayout()
        vbox_right = QVBoxLayout()

        self.create_toolbar()
        self.create_status_bar()

        self.states = States(STATUS_LEDS[self.dct])

        user_positions_widget = self.create_user_position_widget()
        mach_positions_widget = self.create_machine_position_widget()
        actuator_length_widget = self.create_actuator_length_widget()

        vbox_right.addWidget(user_positions_widget)
        vbox_right.addWidget(mach_positions_widget)
        vbox_right.addWidget(actuator_length_widget)

        positions_frame.setLayout(vbox_right)

        vbox_left.addWidget(self.states)

        states_frame.setLayout(vbox_left)

        hbox.addWidget(states_frame)
        hbox.addWidget(positions_frame)

        status_frame.setLayout(hbox)

        tabbed_widget = self.create_tabbed_widget()

        hbox = QHBoxLayout()
        hbox.addWidget(tabbed_widget)
        tabs_frame.setLayout(hbox)

        hbox = QHBoxLayout()
        hbox.addWidget(status_frame)
        hbox.addWidget(tabs_frame)

        app_frame.setLayout(hbox)

        self.setCentralWidget(app_frame)

    def update_status_bar(self, message=None, mode=None, timeout=2000):
        if message:
            self.statusBar().showMessage(message, msecs=timeout)
        if mode:
            self.mode_label.setStyleSheet(f"border: 0; color: {'red' if 'Simulator' in mode else 'black'};")
            self.mode_label.setText(f"mode: {mode}")
        self.statusBar().repaint()

    def updatePositions(self, userPositions, machinePositions, actuatorLengths):
        if userPositions is None:
            MODULE_LOGGER.warning("no userPositions passed into updatePositions(), returning.")
            return

        for upos in range(len(self.user_positions)):
            try:
                self.user_positions[upos][1].setText(f"{userPositions[upos]:10.4f}")
            except IndexError:
                MODULE_LOGGER.error(f"IndexError in user_positions, upos = {upos}")

        if machinePositions is None:
            MODULE_LOGGER.warning("no machinePositions passed into updatePositions(), returning.")
            return

        for mpos in range(len(self.mach_positions)):
            self.mach_positions[mpos][1].setText(f"{machinePositions[mpos]:10.4f}")

        if actuatorLengths is None:
            MODULE_LOGGER.warning("no actuatorLengths passed into updatePositions(), returning.")
            return

        for idx, alen in enumerate(self.actuator_lengths):
            alen[1].setText(f"{actuatorLengths[idx]:10.4f}")

    def updateStates(self, states):
        if states is None:
            return

        self.updateControlButton(states[CONTROL_ONOFF[self.dct]])
        self.states.set_states(states)

    def updateControlButton(self, flag):
        self.control.set_selected(on=flag)


class PunaUIModel(HexapodUIModel):
    def __init__(self, connection_type, device_id):
        hostname, port, dev_id, dev_name, *_ = get_hexapod_controller_pars(device_id)
        if connection_type == "proxy":
            device = ProxyFactory().create(dev_name, device_id=dev_id)
        elif connection_type == "direct":
            device = ControllerFactory().create(dev_name, device_id=dev_id)
            device.connect()
        elif connection_type == "simulator":
            device = PunaSimulator()
        else:
            raise ValueError(f"Unknown type of Hexapod implementation passed into the model: {connection_type}")

        super().__init__(connection_type, device)

        self._device_id = dev_id

        if device is not None:
            MODULE_LOGGER.debug(f"Hexapod initialized as {device.__class__.__name__}")

    @property
    def device_id(self):
        return self._device_id

    def get_device_controller_type(self) -> DCT:
        return DCT.ALPHA_PLUS if isinstance(self.device, (PunaPlusProxy, PunaPlusController)) else DCT.ALPHA

    def get_speed(self):
        vt, vr, vt_min, vr_min, vt_max, vr_max = self.device.get_speed()
        return vt, vr


class PunaUIController(HexapodUIController):
    def __init__(self, model: PunaUIModel, view: PunaUIView):
        super().__init__(model, view)

    def update_values(self):
        super().update_values()

        # Add here any updates to PUNA specific widgets


def parse_arguments():
    """
    Prepare the arguments that are specific for this application.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--type",
        dest="type",
        action="store",
        choices={"proxy", "simulator", "direct"},
        help="Specify Hexapod implementation you want to connect to.",
        default="proxy",
    )
    parser.add_argument(
        "--profile",
        default=False,
        action="store_true",
        help="Enable info logging messages with method profile information.",
    )
    return parser.parse_args()


app = typer.Typer()


@app.command()
def main(device_id: str, device_type: str = "proxy", profile: bool = False):
    multiprocessing.current_process().name = "puna_ui"

    lock_file = QLockFile(str(Path("~/puna_ui.app.lock").expanduser()))

    styles_location = get_resource(":/styles/default.qss")
    app_logo = get_resource(":/icons/logo-puna.svg")

    app = QApplication(["-stylesheet", str(styles_location)])
    app.setWindowIcon(QIcon(str(app_logo)))

    if lock_file.tryLock(100):
        process_status = ProcessStatus()

        timer_thread = threading.Thread(target=do_every, args=(10, process_status.update))
        timer_thread.daemon = True
        timer_thread.start()

        if profile:
            Settings.set_profiling(True)

        if device_type == "proxy":
            _, _, device_id, device_name, *_ = get_hexapod_controller_pars(device_id)
            factory = ProxyFactory()
            proxy = factory.create(device_name, device_id=device_id)
            if not proxy.ping():
                description = "Could not connect to Hexapod Control Server"
                info_text = (
                    "The GUI will start, but the connection button will show a disconnected state. "
                    "Please check if the Control Server is running and start the server if needed. "
                    "Otherwise, check if the correct HOSTNAME for the control server is set in the "
                    "Settings.yaml "
                    "configuration file."
                )

                show_warning_message(description, info_text)

        model = PunaUIModel(device_type, device_id)
        view = PunaUIView(model.get_device_controller_type(), model.device_id)
        PunaUIController(model, view)

        view.show()

        return app.exec_()
    else:
        error_message = QMessageBox()
        error_message.setIcon(QMessageBox.Warning)
        error_message.setWindowTitle("Error")
        error_message.setText("The Puna GUI application is already running!")
        error_message.setStandardButtons(QMessageBox.Ok)

        return error_message.exec()


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG, format=Settings.LOG_FORMAT_FULL)

    sys.exit(app())



"""
The synchronous monitoring service is a small application that performs measurements on
the DAQ6510.

The service reads the configuration for the Keithley DAQ6510 from the
Configuration Manager and then configures the device. When no Configuration Manager is
available, the service can also be started with a filename to read the configuration from. The file
should have the YAML format.

```
insert an excerpt of a sample YAML configuration file here...
```

The monitoring service can be started as follows:


"""

import datetime
import multiprocessing
import os
import sys
import time
from pathlib import Path
from typing import Any

import rich
import typer
from urllib3.exceptions import NewConnectionError

from egse.env import bool_env
from egse.hk import read_conversion_dict
from egse.log import logger
from egse.logger import remote_logging
from egse.metrics import get_metrics_repo
from egse.response import Failure
from egse.scpi import count_number_of_channels, get_channel_names
from egse.settings import get_site_id
from egse.setup import Setup, load_setup
from egse.storage import StorageProxy, is_storage_manager_active
from egse.storage.persistence import CSV
from egse.system import SignalCatcher, flatten_dict, format_datetime, now, str_to_datetime, type_name
from egse.tempcontrol.keithley.daq6510 import DAQ6510Proxy
from egse.tempcontrol.keithley.daq6510_cs import is_daq6510_cs_active

VERBOSE_DEBUG = bool_env("VERBOSE_DEBUG")
SITE_ID = get_site_id()


def load_setup_from_input_file(input_file: str | Path) -> Setup | None:
    """Loads a Setup YAML file from disk."""
    input_file = Path(input_file).resolve()

    if not input_file.exists():
        logger.error(f"ERROR: Input file ({input_file}) doesn't exists.")
        return None

    return Setup.from_yaml_file(input_file)


def daq6510(count, interval, delay, channel_list, input_file: str):
    """
    Run the monitoring service for the DAQ6510.

    Args:
        count: Number of measurements to perform per acquisition [optional]
        interval: Time interval between measurements in seconds [optional]
        delay: Delay between acquisitions in seconds [optional]
        channel_list: Comma-separated list of channels to acquire data from [optional]
        input_file: YAML file containing the Setup for the DAQ6510 [optional]

    """

    if input_file:
        setup = load_setup_from_input_file(input_file)
    else:
        setup = load_setup()

    if setup is None:
        logger.error("ERROR: Could not load setup.")
        sys.exit(1)

    if VERBOSE_DEBUG:
        logger.debug(f"Loaded setup: {setup}")

    if not hasattr(setup, "gse"):
        logger.error("ERROR: No GSE section in the loaded Setup.")
        sys.exit(1)

    try:
        hk_conversion_table = read_conversion_dict("DAQ6510-MON", use_site=True, setup=setup)
        column_names = list(hk_conversion_table.values())
    except Exception as exc:
        logger.warning(f"WARNING: Failed to read telemetry dictionary: {exc}")
        hk_conversion_table = {"101": "PT100-4", "102": "PT100-2"}
        column_names = list(hk_conversion_table.values())

    if not is_daq6510_cs_active():
        logger.error(
            "The DAQ6510 Control Server is not running, start the 'daq6510_cs' command "
            "before running the data acquisition."
        )
        return

    if not is_storage_manager_active():
        logger.error("The storage manager is not running, start the core services before running the data acquisition.")
        return

    if "DAQ6510" not in setup.gse:  # type: ignore
        logger.error("ERROR: no DAQ6510 entry in the loaded Setup.")
        sys.exit(1)

    if not channel_list:
        channel_list = setup.gse.DAQ6510.channels  # type: ignore

    if not count:
        count = setup.gse.DAQ6510.route.scan.count.scan  # type: ignore

    if not interval:
        interval = setup.gse.DAQ6510.route.scan.interval  # type: ignore

    if not delay:
        delay = setup.gse.DAQ6510.route.delay  # type: ignore

    count, interval, delay = int(count), int(interval), int(delay)

    channel_count = count_number_of_channels(channel_list)
    channel_names = get_channel_names(channel_list)

    metrics_client = setup_metrics_client()

    # Initialize some variables that will be used for registration to the Storage Manager

    origin = "DAQ6510-MON"
    persistence_class = CSV
    prep = {
        "mode": "a",
        "ending": "\n",
        "column_names": ["timestamp", *column_names],
    }

    killer = SignalCatcher()

    with DAQ6510Proxy() as daq, StorageProxy() as storage:
        daq.reset()

        dt = now()
        daq.set_time(dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second)
        logger.info(f"DAQ6510 date and time set: {daq.get_time()}")

        storage.register({"origin": origin, "persistence_class": persistence_class, "prep": prep})

        # This will write a comment line to the CSV file with the column names. This might be useful when
        # the sensors are reconfigured and the number or names of columns changes.
        storage.save({"origin": origin, "data": f"# columns: {column_names}"})

        for sensor in setup.gse.DAQ6510.sensors:  # type: ignore
            for function in setup.gse.DAQ6510.sensors[sensor]:  # type: ignore
                sense = {
                    function.upper(): [
                        (key, value)
                        for key, value in flatten_dict(setup.gse.DAQ6510.sensors[sensor][function]).items()  # type: ignore
                        if key != "channels"
                    ]
                }
                function_channel_list = setup.gse.DAQ6510.sensors[sensor][function].channels  # type: ignore
                if VERBOSE_DEBUG:
                    logger.debug(f"{sense=}")
                    logger.debug(f"{function_channel_list=}")
                daq.configure_sensors(channel_list=function_channel_list, sense=sense)

        logger.info(f"global: {channel_list=}, {channel_count=}")

        daq.setup_measurements(channel_list=channel_list)

        while True:
            try:
                response = daq.perform_measurement(channel_list=channel_list, count=count, interval=interval)

                if killer.term_signal_received:
                    break

                if not response:
                    logger.warning("Received an empty response from the DAQ6510, check the connection with the device.")
                    logger.warning(f"Response: {response=}")
                    time.sleep(1.0)
                    continue

                if isinstance(response, Failure):
                    logger.warning("Received a Failure from the DAQ6510 Control Server:")
                    logger.warning(f"Response: {response}")
                    time.sleep(1.0)
                    continue

                # Process and save the response

                if VERBOSE_DEBUG:
                    logger.debug(f"{response=}")

                dts = response[0][1].strip()
                dt = datetime.datetime.strptime(dts[:-3], "%m/%d/%Y %H:%M:%S.%f")
                datetime_string = format_datetime(dt.replace(tzinfo=datetime.timezone.utc))

                data: dict[str, Any] = {hk_conversion_table[measure[0]]: float(measure[2]) for measure in response}
                data.update({"timestamp": datetime_string})

                # FIXME: we probably need to do something with the units...

                units = [measure[3] for measure in response]

                if VERBOSE_DEBUG:
                    logger.debug(f"{data=}")

                storage.save({"origin": origin, "data": data})

                # Now extract channels from the response to update the metrics

                for channel in [measure[0] for measure in response]:
                    if channel in hk_conversion_table:
                        metrics_name = hk_conversion_table[channel]
                        save_metrics(metrics_client, origin, data)

                # wait for the next measurement to be done (delay)

                time.sleep(delay)

            except KeyboardInterrupt:
                logger.debug("Interrupt received, terminating...")
                break
            except Exception as exc:
                logger.warning(f"{type_name(exc)}: {exc}", exc_info=True)
                logger.warning("Got a corrupt response from the DAQ6510. Check log messages for 'DAS Exception'.")
                time.sleep(1.0)
                continue

        storage.unregister({"origin": origin})

    logger.info("DAQ6510 Data Acquisition System terminated.")


def setup_metrics_client():
    token = os.getenv("INFLUXDB3_AUTH_TOKEN")
    project = os.getenv("PROJECT")

    if project and token:
        metrics_client = get_metrics_repo(
            "influxdb", {"host": "http://localhost:8181", "database": project, "token": token}
        )
        metrics_client.connect()
    else:
        metrics_client = None
        logger.warning(
            "INFLUXDB3_AUTH_TOKEN and/or PROJECT environment variable is not set. "
            "Metrics will not be propagated to InfluxDB."
        )

    return metrics_client


def save_metrics(metrics_client, origin, data):
    try:
        if metrics_client:
            point = {
                "measurement": origin.lower(),
                "tags": {"site_id": SITE_ID, "origin": origin},
                "fields": {hk_name.lower(): data[hk_name] for hk_name in data if hk_name != "timestamp"},
                "time": str_to_datetime(data["timestamp"]),
            }
            metrics_client.write(point)
        else:
            logger.warning(
                f"Could not write {origin} metrics to the time series database (self.metrics_client is None)."
            )
    except NewConnectionError:
        logger.warning(
            f"No connection to the time series database could be established to propagate {origin} metrics.  Check "
            f"whether this service is (still) running."
        )


app = typer.Typer(
    name="daq6510_mon",
    help="DAQ6510 Data Acquisition Unit, Keithley, temperature monitoring (monitoring)",
    no_args_is_help=True,
)


@app.command()
def start(input_file: str = typer.Option("", help="YAML file containing the Setup for the DAQ6510")):
    """Starts the Keithley DAQ6510 Monitoring Service."""

    multiprocessing.current_process().name = "daq6510_mon (start)"

    with remote_logging():
        from egse.env import setup_env

        setup_env()

        try:
            daq6510(count=None, interval=None, delay=None, channel_list=None, input_file=input_file)
        except KeyboardInterrupt:
            logger.debug("Shutdown requested...exiting")
        except SystemExit as exit_code:
            logger.debug("System Exit with code {}.".format(exit_code))
            sys.exit(exit_code.code)
        except Exception:
            msg = "Cannot start the DAQ6510 Monitoring Service"
            logger.exception(msg)
            rich.print(f"[red]{msg}.")


if __name__ == "__main__":
    sys.exit(app())



from __future__ import annotations

import importlib
import multiprocessing
import pickle
import time
from typing import Callable
from typing import Tuple

import typer
import zmq
from zmq import ZMQError

from egse.control import ControlServer
from egse.log import logger
from egse.protocol import CommandProtocol
from egse.settings import Settings
from egse.setup import load_setup
from egse.system import format_datetime
from egse.zmq_ser import MessageIdentifier
from egse.zmq_ser import bind_address


class MonitoringProtocol(CommandProtocol):
    def __init__(self, control_server: ControlServer):
        super().__init__(control_server)

    def get_bind_address(self):
        return bind_address(self.control_server.get_communication_protocol(), self.control_server.get_monitoring_port())

    def get_status(self):
        return {
            "timestamp": format_datetime(),
        }

    def send_status(self, status):
        self.send(status)

    def get_housekeeping(self) -> dict:
        return {
            "timestamp": format_datetime(),
        }


class Monitoring:
    """Context manager to monitor processes.

    Most control servers publish status information on their `MONITORING_PORT`.

    Parameters:
        endpoint: the endpoint to which the service will connect
        subscribe: subscription string, default is 'ALL'
        use_pickle: use pickle to process responses, currently always True
        callback: function that is called to process the response
        timeout: stop monitoring after timeout seconds
    """

    def __init__(
        self,
        endpoint: str,
        subscribe: Tuple[str] = None,
        use_pickle: bool = True,
        callback: Callable = None,
        timeout: float = None,  # in fractional seconds
    ):
        self._subscribe = subscribe or ["ALL"]
        self._endpoint = endpoint
        self._context = zmq.Context().instance()
        self._socket = None
        self._setup = None
        self._subscriptions = set()
        self._use_pickle = use_pickle
        self._callback = callback
        self._timeout = timeout
        self.return_code = None

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self._socket.closed:
            self.disconnect()

    def connect(self):
        logger.debug(f"Connecting to {self._endpoint}")
        self._socket = self._context.socket(zmq.SUB)
        self._socket.connect(self._endpoint)

        # Get the high watermark value for the reception socket
        rcv_hwm = self._socket.getsockopt(zmq.RCVHWM)

        logger.debug(f"Receive High Water Mark: {rcv_hwm}")

        self._setup = load_setup()

        logger.debug(f"Using Setup {self._setup.get_id()}")

    def disconnect(self):
        logger.debug(f"Disconnecting from {self._endpoint}")
        self._socket.close(linger=0)
        self._subscriptions.clear()

    def unsubscribe_all(self):
        for sub in self._subscriptions:
            self._socket.unsubscribe(sub)
        self._subscriptions.clear()

    def unsubscribe(self, sync_id: int):
        subscribe_string = sync_id.to_bytes(1, byteorder="big") if sync_id else b""
        try:
            self._subscriptions.remove(subscribe_string)
            self._socket.unsubscribe(subscribe_string)
        except KeyError:
            logger.warning(f"Trying to unsubscribe a key that was not previously subscribed: {subscribe_string}")

    def subscribe(self, sync_id: int = None):
        subscribe_string = sync_id.to_bytes(1, byteorder="big") if sync_id else b""

        logger.debug(f"Subscribing {sync_id}")

        if subscribe_string in self._subscriptions:
            return

        self._socket.subscribe(subscribe_string)
        self._subscriptions.add(subscribe_string)

    def _clear_message_queue(self):
        # unsubscribing and subscribing again doesn't seem to work, so we close and re-open the socket.
        # Then we subscribe to restore the previous state.

        self._socket.close(linger=0)

        self._socket = self._context.socket(zmq.SUB)
        self._socket.connect(self._endpoint)

        for sub in self._subscriptions:
            self._socket.subscribe(sub)

    def handle_multi_part(self, message_id) -> Tuple[int, list]:
        message_parts = []
        message_id = int.from_bytes(message_id, byteorder="big")
        while True:
            message_parts.append(pickle.loads(self._socket.recv()))
            if not self._socket.getsockopt(zmq.RCVMORE):
                break

        return message_id, message_parts

    def handle_single_part(self, message) -> Tuple[int, list]:
        message_id = MessageIdentifier.ALL
        response = pickle.loads(message)

        return message_id, [response]

    def run(self):
        for item in self._subscribe or ["ALL"]:
            try:
                sync_id = MessageIdentifier[item.upper()]
            except KeyError:
                msg = f"incorrect subscribe identifier, use one of {[x.name for x in MessageIdentifier]}"
                logger.error(f"[red]ERROR: {msg}[/]")
                raise ValueError(msg)
            else:
                self.subscribe(sync_id)

        return self._monitoring_loop()

    def _monitoring_loop(self):
        start_time = time.monotonic()

        while True:
            try:
                message = self._socket.recv(zmq.NOBLOCK)
                more_parts = self._socket.getsockopt(zmq.RCVMORE)

                if more_parts:
                    sync_id, response = self.handle_multi_part(message)
                else:
                    sync_id, response = self.handle_single_part(message)

                start_time = time.monotonic()

                if self._callback:
                    rc = self._callback(sync_id, response, self._setup)
                    if isinstance(rc, (tuple, list)):
                        do_break, rc, *_ = rc  # discard any extra values
                    else:
                        do_break = rc
                    self.return_code = rc
                    if do_break:
                        break
                else:
                    logger.info(f"{MessageIdentifier(sync_id).name}, {response}")
            except ZMQError:
                if self._timeout and time.monotonic() - start_time > self._timeout:
                    self.return_code = None
                    break
            except KeyboardInterrupt:
                logger.info("KeyboardInterrupt caught!")
                break
            except pickle.UnpicklingError as exc:
                logger.error(f"UnpicklingError: {exc}")

        return self.return_code

    def resume(self, callback=None, clear_messages: bool = False):
        if clear_messages:
            self._clear_message_queue()
        if callback:
            self._callback = callback

        return self._monitoring_loop()

    def get_return_code(self):
        return self.return_code


PROCESS_NAMES = {
    "DATA_DUMP": ("DATA DUMPER", "MONITORING_PORT"),
    "DPU_PROCESSOR": ("DPU Processor", "MONITORING_PORT"),
    "DPU_PROCESSOR_HEARTBEAT": ("DPU Processor", "HEARTBEAT_PORT"),
    "DPU_CS": ("DPU Control Server", "MONITORING_PORT"),
    "CM_CS": ("Configuration Manager Control Server", "MONITORING_PORT"),
    "SM_CS": ("Storage Manager Control Server", "MONITORING_PORT"),
    "PM_CS": ("Process Manager Control Server", "MONITORING_PORT"),
    "SYN_CS": ("Synoptics Manager Control Server", "MONITORING_PORT"),
    "DATA_DISTRIBUTION": ("DPU Processor", "DATA_DISTRIBUTION_PORT"),
}


def _determine_port(proc_name: str):
    """
    Determine the port number for the given process name. The process names are specific for this tool and defined
    in the PROCESS_NAMES dictionary.
    """
    name, port = PROCESS_NAMES.get(proc_name.upper(), (None, None))

    logger.info(f"{name = }, {port = }")

    if name is not None:
        ctrl_settings = Settings.load(name)
        port = ctrl_settings.get(port)

    return port


app = typer.Typer()


@app.command()
def monitoring(
    hostname: str,
    port: str,
    subscribe: Tuple[str] = None,
    use_pickle: bool = True,
    list_names: bool = False,
    callback: str = None,
):
    """
    Monitor the status of a control server on hostname:port.

    The port number shall correspond to the port number on which the control server is publishing
    status information.
    """

    multiprocessing.current_process().name = "Monitoring"

    from rich import print
    from egse.env import setup_env

    setup_env()

    if list_names:
        print("The available process names as aliases for port numbers are: ", end="")
        print(", ".join(PROCESS_NAMES))
        return
    elif hostname is None or port is None:
        print("You will need to provide both the HOSTNAME and the PORT argument.")
        return

    logger.info(f"{type(hostname) = }, {hostname = }")
    logger.info(f"{type(port) = }, {port = }, {_determine_port(port) = }")

    try:
        port = int(port)
    except ValueError:
        port = _determine_port(port_name := port)
        if port is None:
            print(
                f"[red]ERROR[/]: Couldn't determine port number from {port_name}, "
                f"use the '--list-names' flag to see available names."
            )
            return

    if callback is not None:
        callback_module = importlib.import_module("egse.monitoring")
        callback = getattr(callback_module, callback, None)

    with Monitoring(f"tcp://{hostname}:{port}", subscribe=subscribe, use_pickle=use_pickle, callback=callback) as moni:
        moni.run()


if __name__ == "__main__":
    app()



import logging

from egse.arbitrary_wave_generator.aim_tti import DEVICE_SETTINGS

LOGGER = logging.getLogger(__name__)

IDENTIFICATION_QUERY = "*IDN?"


import logging
import socket
import time
from string import digits


from egse.device import (
    DeviceConnectionError,
    DeviceTimeoutError,
    DeviceError,
    DeviceConnectionInterface,
    DeviceTransport,
)

logger = logging.getLogger(__name__)

CONNECT_TIMEOUT = 3.0  # Timeout when connecting the socket [s]

remove_digits = str.maketrans("", "", digits)
time_in_s = time.time()


class TgfError(Exception):
    """A TGF-specific error."""

    pass


class Tgf4000EthernetInterface(DeviceConnectionInterface, DeviceTransport):
    """Ethernet Interface for the TGF devices."""

    def __init__(self, hostname: str = None, port: int = None, device_id: str = None, read_timeout: float = 60):
        """Initialisation of an Ethernet interface for a TGF device.

        Args:
            hostname( str): Hostname to which to open a socket.
            port (int): Port to which to open a socket.
            device_id (str): Identifier of the device to which to open a socket.
            read_timeout (float): Timeout for reading commands [s].
        """

        super().__init__()

        print(f"Device ID: {device_id}")

        self.hostname = DEVICE_SETTINGS[device_id]["HOSTNAME"] if hostname is None else hostname
        self.port = DEVICE_SETTINGS[device_id]["PORT"] if port is None else port
        self.device_id = device_id
        self._sock = None

        print(f"Hostname: {self.hostname}")
        print(f"Port: {self.port}")
        print(f"Device ID: {self.device_id}")

        self._is_connection_open = False
        self.read_timeout = read_timeout

    def connect(self) -> None:
        """Connects to the Aim-TTi TGF4000 hardware.

        Raises:
            DeviceConnectionError: When the connection could not be established. Check the logging messages for more
                                   details.
            DeviceTimeoutError: When the connection timed out.
            ValueError: When hostname or port number are not provided.
        """

        # Sanity checks

        if self._is_connection_open:
            logger.warning(f"{self.device_id}: trying to connect to an already connected socket.")
            return

        if self.hostname in (None, ""):
            raise ValueError(f"{self.device_id}: hostname is not initialised.")

        if self.port in (None, 0):
            raise ValueError(f"{self.device_id}: port number is not initialised.")

        # Create a new socket instance

        try:
            self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            # The following lines are to experiment with blocking and timeout, but there is no need.
            # self._sock.setblocking(1)
            # self._sock.settimeout(3)
        except socket.error as e_socket:
            raise DeviceConnectionError(self.device_id, "Failed to create socket.") from e_socket

        print(f"Socket created: {self._sock}")

        # Attempt to establish a connection to the remote host

        # FIXME: Socket shall be closed on exception?

        # We set a timeout of 3s before connecting and reset to None (=blocking) after the `connect` method has been
        # called. This is because when no device is available, e.g. during testing, the timeout will take about
        # two minutes, which is way too long. It needs to be evaluated if this approach is acceptable and not causing
        # problems during production.

        try:
            logger.debug(f'Connecting a socket to host "{self.hostname}" using port {self.port}')
            self._sock.settimeout(3)
            self._sock.connect((self.hostname, self.port))
            self._sock.settimeout(None)
        except ConnectionRefusedError as exc:
            raise DeviceConnectionError(self.device_id, f"Connection refused to {self.hostname}:{self.port}.") from exc
        except TimeoutError as exc:
            raise DeviceTimeoutError(self.device_id, f"Connection to {self.hostname}:{self.port} timed out.") from exc
        except socket.gaierror as exc:
            raise DeviceConnectionError(self.device_id, f"Socket address info error for {self.hostname}") from exc
        except socket.herror as exc:
            raise DeviceConnectionError(self.device_id, f"Socket host address error for {self.hostname}") from exc
        except OSError as exc:
            raise DeviceConnectionError(self.device_id, f"OSError caught ({exc}).") from exc

        self._is_connection_open = True

        # Check that we are connected to the controller by issuing the "VERSION" or
        # "*IDN?" query. If we don't get the right response, then disconnect automatically.

        if not self.is_connected():
            raise DeviceConnectionError(
                self.device_id, "Device is not connected, check logging messages for the cause."
            )

    def disconnect(self) -> None:
        """Disconnects from the Aim-TTi TGF4000 hardware.

        Raises:
            DeviceConnectionError when the socket could not be closed.
        """

        try:
            if self._is_connection_open:
                logger.debug(f"Disconnecting from {self.hostname}")
                self._sock.close()
                self._is_connection_open = False
        except Exception as e_exc:
            raise DeviceConnectionError(self.device_id, f"Could not close socket to {self.hostname}") from e_exc

    def reconnect(self):
        """Reconnects to the Aim-TTi TGF4000 hardware.

        Raises:
            ConnectionError when the device cannot be reconnected for some reason.
        """

        if self._is_connection_open:
            self.disconnect()
        self.connect()

    def is_connected(self) -> bool:
        """Checks if the Aim-TTi TGF4000 hardware is connected.

        This will send a query for the device identification and validate the answer.

        Returns: True is the device is connected and answered with the proper ID; False otherwise.
        """

        if not self._is_connection_open:
            return False

        try:
            print(f"Result from identification query: {self.query(IDENTIFICATION_QUERY)}")
            # noinspection PyTypeChecker
            manufacturer, model, *_ = self.query(IDENTIFICATION_QUERY).split(",")

        except DeviceError as exc:
            logger.exception(exc)
            logger.error("Most probably the client connection was closed. Disconnecting...")
            self.disconnect()
            return False

        print(f"Manufacturer: {manufacturer}")

        if "THURLBY THANDAR" not in manufacturer or "TGF4" not in model:
            logger.error(
                f"Device did not respond correctly to a {IDENTIFICATION_QUERY} command, manufacturer={manufacturer}, "
                f"model={model}. Disconnecting..."
            )
            self.disconnect()
            return False

        return True

    def write(self, command: str) -> None:
        """Sends a single command to the device controller without waiting for a response.

        Args:
            command (str): Command to send to the controller

        Raises:
            DeviceConnectionError when the command could not be sent due to a communication problem.
            DeviceTimeoutError when the command could not be sent due to a timeout.
        """

        try:
            command += "\n" if not command.endswith("\n") else ""

            self._sock.sendall(command.encode())

        except socket.timeout as e_timeout:
            raise DeviceTimeoutError(self.device_id, "Socket timeout error") from e_timeout
        except socket.error as e_socket:
            # Interpret any socket-related error as a connection error
            raise DeviceConnectionError(self.device_id, "Socket communication error.") from e_socket
        except AttributeError:
            if not self._is_connection_open:
                msg = "The TGF4000 is not connected, use the connect() method."
                raise DeviceConnectionError(self.device_id, msg)
            raise

    def trans(self, command: str) -> str | bytes:
        """Sends a single command to the device controller and block until a response from the controller.

        This is seen as a transaction.

        Args:
            command (str): Command to send to the controller

        Returns:
            Either a string returned by the controller (on success), or an error message (on failure).

        Raises:
            DeviceConnectionError when there was an I/O problem during communication with the controller.
            DeviceTimeoutError when there was a timeout in either sending the command or receiving the response.
        """

        try:
            # Attempt to send the complete command

            command += "\n" if not command.endswith("\n") else ""

            self._sock.sendall(command.encode())

            # wait for, read and return the response from Aim-TTi TGF4000 (will be at most TBD chars)

            return_string = self.read()

            return return_string.decode().rstrip()

        except UnicodeError:
            # noinspection PyUnboundLocalVariable
            return return_string
        except socket.timeout as e_timeout:
            raise DeviceTimeoutError(self.device_id, "Socket timeout error") from e_timeout
        except socket.error as e_socket:
            # Interpret any socket-related error as an I/O error
            raise DeviceConnectionError(self.device_id, "Socket communication error.") from e_socket
        except ConnectionError as exc:
            raise DeviceConnectionError(self.device_id, "Connection error.") from exc
        except AttributeError:
            if not self._is_connection_open:
                raise DeviceConnectionError(self.device_id, "Device not connected, use the connect() method.")
            raise

    def read(self) -> bytes:
        """Reads from the device buffer.

        Returns: Content of the device buffer.
        """

        n_total = 0
        buf_size = 2048

        # Set a timeout of READ_TIMEOUT to the socket.recv

        saved_timeout = self._sock.gettimeout()
        self._sock.settimeout(self.read_timeout)

        try:
            for idx in range(100):
                time.sleep(0.001)  # Give the device time to fill the buffer
                data = self._sock.recv(buf_size)
                n = len(data)
                n_total += n
                if n < buf_size:
                    break
        except socket.timeout:
            logger.warning(f"Socket timeout error for {self.hostname}:{self.port}")
            return b"\r\n"
        except TimeoutError as exc:
            logger.warning(f"Socket timeout error: {exc}")
            return b"\r\n"
        finally:
            self._sock.settimeout(saved_timeout)

        # noinspection PyUnboundLocalVariable
        return data


def main():
    return 0


if __name__ == "__main__":
    main()



from egse.digilent.measurpoint.digilent_devif import DigilentEthernetInterface
from egse.settings import Settings

DEVICE_SETTINGS = Settings.load("Digilent MEASURpoint DT8874")


class Dt8874EthernetInterface(DigilentEthernetInterface):
    def __init__(self, hostname: str = None, port: int = None):
        hostname = DEVICE_SETTINGS.HOSTNAME if hostname is None else hostname
        port = DEVICE_SETTINGS.PORT if port is None else port
        device_name = DEVICE_SETTINGS.DEVICE_NAME
        timeout = DEVICE_SETTINGS.TIMEOUT

        super().__init__(hostname, port, device_name, timeout)



"""
Tests for the egse.version module.
"""

import subprocess
import textwrap
from unittest.mock import MagicMock
from unittest.mock import patch

import pytest

from egse.version import get_version_from_git
from egse.version import get_version_from_settings
from egse.version import get_version_from_settings_file_raw
from egse.version import get_version_installed


@patch("egse.plugin.entry_points")
@patch("egse.version.get_version_installed")
@patch("egse.version.get_version_from_git")
def test_main_execution(mock_git, mock_installed, mock_entry_points, capsys):
    """Test the main script execution."""
    mock_git.return_value = "2024.2.0-5-g1234567"
    mock_installed.return_value = "2024.1.0"
    mock_entry_point = MagicMock()
    mock_entry_point.name = "test-package"
    mock_entry_points.return_value = [mock_entry_point]

    # Import, execute the main function, and capture the output

    import egse.version
    from egse.version import main

    main()

    captured = capsys.readouterr()

    assert "CGSE version in Settings:" in captured.out
    assert egse.version.VERSION in captured.out

    assert "git version (current project) =" in captured.out
    assert "2024.2.0-5-g1234567" in captured.out

    assert "Installed version for test-package=" in captured.out
    assert "2024.1.0" in captured.out

    assert hasattr(egse.version, "VERSION")
    assert hasattr(egse.version, "__PYPI_VERSION__")

    captured = capsys.readouterr()

    with patch("egse.version.VERSION", None):
        main()
        assert "CGSE version in Settings:" not in captured.out


def test_get_version_from_settings():
    """Test reading version from settings.yaml file or a YAML string."""

    settings_content = textwrap.dedent(
        """
        CGSE:
            VERSION: 2025.11.25
            # This is a comment
            SITE_ID: TEST_LAB
        """
    )

    # Test with yaml_string input, this will take precedence over file input.

    version = get_version_from_settings("CGSE", yaml_string=settings_content)
    assert version == "2025.11.25"

    # Test mocking Settings.load to return specific values. We need to patch.object here since
    # Settings is imported inside the function `get_version_from_settings()`.

    import egse.settings
    from egse.settings import SettingsError
    from egse.system import attrdict

    with patch.object(egse.settings, "Settings", autospec=True) as inner_settings:
        inner_settings.load.return_value = attrdict({"VERSION": "2025.11.26", "SITE_ID": "TEST_LAB_2"})
        version = get_version_from_settings("CGSE")
        assert version == "2025.11.26"

    # Test with incorrect group name, this will result in eventually calling get_version_from_settings_file_raw()
    # which is tested below. With the WRONG_GROUP name, Settings.load() will raise a SettingsError
    # and the get_version_from_settings_file_raw() will raise a RuntimeError.

    with patch.object(egse.settings, "Settings", autospec=True) as inner_settings:
        inner_settings.load.side_effect = SettingsError
        with pytest.raises(RuntimeError):
            get_version_from_settings("WRONG_GROUP")


def test_valid_settings_file(tmp_path):
    """Test reading version from a valid settings.yaml file."""
    settings_file = tmp_path / "settings.yaml"
    settings_content = textwrap.dedent(
        """\
        CGSE:
            VERSION: 2024.1.23
            # This is a comment
            OTHER_SETTING: value
        """
    )
    settings_file.write_text(settings_content)

    version = get_version_from_settings_file_raw("CGSE", location=tmp_path)
    assert version == "2024.1.23"


def test_version_with_inline_comment(tmp_path):
    """Test reading version with inline comment."""
    settings_file = tmp_path / "settings.yaml"
    settings_content = textwrap.dedent(
        """\
        CGSE:
            VERSION: 2024.1.24  # inline comment
        """
    )
    settings_file.write_text(settings_content)

    version = get_version_from_settings_file_raw("CGSE", location=tmp_path)
    assert version == "2024.1.24"


def test_invalid_group_name(tmp_path):
    """Test error when group name doesn't match."""
    settings_file = tmp_path / "settings.yaml"
    settings_content = textwrap.dedent(
        """\
        CGSE:
            VERSION: 2024.1.0
        """
    )
    settings_file.write_text(settings_content)

    with pytest.raises(RuntimeError, match="Incompatible format.*should start with 'WRONG_GROUP'"):
        get_version_from_settings_file_raw("WRONG_GROUP", location=tmp_path)


def test_missing_version_field(tmp_path):
    """Test error when VERSION field is missing."""
    settings_file = tmp_path / "settings.yaml"
    settings_content = textwrap.dedent(
        """\
        CGSE:
            OTHER_FIELD: random value
        """
    )
    settings_file.write_text(settings_content)

    with pytest.raises(RuntimeError, match="no VERSION found"):
        get_version_from_settings_file_raw("CGSE", location=tmp_path)


@patch("egse.system.chdir")
@patch("subprocess.run")
def test_successful_git_describe(mock_run, mock_chdir):
    """Test successful git describe command."""
    mock_proc = MagicMock()
    mock_proc.stdout = b"2024.1.0-5-g1234567\n"
    mock_proc.stderr = b""
    mock_run.return_value = mock_proc

    version = get_version_from_git()
    assert version == "2024.1.0-5-g1234567"
    mock_run.assert_called_once_with(
        ["git", "describe", "--tags", "--long", "--always"],
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
    )


@patch("egse.system.chdir")
@patch("subprocess.run")
def test_empty_stdout(mock_run, mock_chdir):
    """Test when git command returns empty stdout."""
    mock_proc = MagicMock()
    mock_proc.stdout = b""
    mock_proc.stderr = b"some error"
    mock_run.return_value = mock_proc

    version = get_version_from_git()

    assert version == "0.0.0"


@patch("egse.system.chdir")
@patch("subprocess.run")
def test_git_command_failure(mock_run, mock_chdir):
    """Test when git command raises CalledProcessError."""
    mock_run.side_effect = subprocess.CalledProcessError(1, "git")

    version = get_version_from_git()

    assert version == "0.0.0"


@patch("egse.system.chdir")
@patch("subprocess.run")
def test_with_specific_location(mock_run, mock_chdir, tmp_path):
    """Test git describe with specific location."""
    mock_proc = MagicMock()
    mock_proc.stdout = b"2024.2.0-0-gabcdefg\n"
    mock_proc.stderr = b""
    mock_run.return_value = mock_proc

    version = get_version_from_git(location=tmp_path)
    assert version == "2024.2.0-0-gabcdefg"


# Tests for get_version_installed function.


@patch("importlib.metadata.version")
def test_package_found(mock_get_version):
    """Test when package is found in metadata."""
    mock_get_version.return_value = "2024.1.0+local.123"

    version = get_version_installed("cgse-common")
    assert version == "2024.1.0+local.123"


@patch("importlib.metadata.version")
def test_package_not_found(mock_get_version):
    """Test when package is not found in metadata."""
    from importlib.metadata import PackageNotFoundError

    mock_get_version.side_effect = PackageNotFoundError("cgse-unknown")

    version = get_version_installed("cgse-unknown")

    assert version == "0.0.0"


# Integration tests for version module variables.


def test_version_is_set():
    """Test that VERSION variable is set."""
    from egse.version import VERSION

    assert VERSION is not None
    # VERSION should either be a valid version string or "0.0.0"
    assert isinstance(VERSION, str)
    assert len(VERSION) > 0


def test_pypi_version_format():
    """Test that __PYPI_VERSION__ is properly formatted."""
    from egse.version import __PYPI_VERSION__

    assert __PYPI_VERSION__ is not None
    assert isinstance(__PYPI_VERSION__, str)
    # Should not contain '+' as it's split on that character
    assert "+" not in __PYPI_VERSION__


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



"""General information for the Ariel Telescope Control Unit (TCU).

Reference documents:
    - RD01: TCU User Manual (ARIEL-IEEC-PL-TN-002), v1.2
    - RD02: ARIEL TCU Data Handling (ARIEL-IEEC-PL-TN-007), v1.0
    - RD03: TCU code provided by Vladimiro Noce (priv. comm.)
    - RD04: ARIEL Telescope Control Unit Design Description Document (ARIEL-IEEC-PL-DD-001), v1.10
    - RD05: ARIEL TCU FW Architecture Design(ARIEL-IEEC-PL-DD-002), v1.5
"""

from enum import IntEnum
from pathlib import Path

from egse.settings import Settings

HERE = Path(__file__).parent
settings = Settings.load("Ariel TCU Control Server")

# General information about the TCU Control Server

PROCESS_NAME = settings.get("PROCESS_NAME", "tcu_cs")  # Name under which it is registered in the service registry
SERVICE_TYPE = settings.get(
    "SERVICE_TYPE", "tcu_cs"
)  # Service type under which it is registered in the service registry
PROTOCOL = settings.get("PROTOCOL", "tcp")  # Communication protocol
HOSTNAME = settings.get("HOSTNAME", "localhost")  # Hostname
COMMANDING_PORT = settings.get("COMMANDING_PORT", 0)  # Commanding port (as per settings or dynamically assigned)
SERVICE_PORT = settings.get("SERVICE_PORT", 0)  # Service port (as per settings or dynamically assigned)
MONITORING_PORT = settings.get("MONITORING_PORT", 0)  # Monitoring port (as per settings or dynamically assigned)
STORAGE_MNEMONIC = settings.get("STORAGE_MNEMONIC", "TCU")  # Storage mnemonic (used in the HK filenames)

NUM_TSM_FRAMES = settings.get("NUM_TSM_FRAMES", 2)
NUM_TSM_PROBES_PER_FRAME = settings.get("NUM_TSM_PROBES_PER_FRAME", 22)
NUM_M2MD_POSITIONS = settings.get("NUM_M2MD_POSITIONS", 18)
NUM_M2MD_AXES = settings.get("NUM_M2MD_AXES", 3)

PROXY_TIMEOUT = 10


class TcuMode(IntEnum):
    """Ariel TCU operating modes.

    The different TCU modes are:
    - IDLE: Waiting for commands, minimum power consumption,
    - BASE: HK + TSM circuitry on,
    - CALIBRATION: HK + TSM + M2MD circuitry on.
    """

    # Adopted from Vladimiro's code

    IDLE = 0x0000  # Waiting for commands, minimum power consumption
    BASE = 0x0001  # HK + TSM circuitry on
    CALIBRATION = 0x0003  # HK + TSM + M2MD circuitry on


class MotorState(IntEnum):
    """State of the M2MD motors.

    The different motor states are:
    - STANDBY: No motion,
    - OPERATION: Motor moving.
    """

    # Adopted from Vladimiro's code
    # RD01 -> Sect. 5.1

    STANDBY = 0x0001  # No motion
    OPERATION = 0x0010  # Motor moving


# Axis velocity settings
#   - Keys: Axis speed in Hz
#   - Values: Hex string representing the axis speed
#             -> This has to be used as cargo2 in the hex command that is sent to the Arduino
AXIS_VELOCITY = {
    1: 0x5DC0,  # Writing @ 1Hz
    2: 0x2EE0,  # Writing @ 2Hz
    4: 0x1770,  # Writing @ 4Hz
    8: 0x0BB8,  # Writing @ 8Hz
    16: 0x05DC,  # Writing @ 16Hz
    32: 0x02EE,  # Writing @ 32Hz
    64: 0x0177,  # Writing @ 64Hz
    # For testing purposes
    75: 0x0140,  # Writing @ 75Hz
    80: 0x012C,  # Writing @ 80Hz
    96: 0x00FA,  # Writing @ 96Hz
}


# To interpret the error code values
# Taken from RD02 -> Table 3

ERROR_CODES = {
    0x00: "No errors",
    0x01: "Command not executed",
    0x02: "Command error",
    0x03: "Command not allowed",
    0x04: "Destination address error",
    0x05: "Protocol ID error",
    0x06: "Source address error",
    0x07: "Unexpected sequence error",
    0x08: "Address error",
    0x09: "Data length error",
    0x0A: "CRC error",
    0x0B: "Parameter values error",
    0x0C: "Bad mode",
    0x0D: "Timeout",
    0x0E: "Unknown",
    0x0F: "Extended error, see cargo",
}
